/*********************************************************************
 *
 *      Copyright (C) 1999-2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      JSwat
 * FILE:        BasicSession.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      04/09/99        Initial version
 *      nf      08/10/00        Removed the debuggee connecting code
 *                              and put it in the commands & actions
 *      nf      03/11/01        Added support for properties
 *      nf      04/24/01        Fixed bug 104, suspend VM on start
 *      nf      05/19/01        Fixed bugs 64 and 113
 *      nf      05/28/01        Fixed bug 116
 *      nf      06/17/01        Use the new breakpoint manager to catch
 *                              uncaught exceptions
 *      nf      07/21/01        Added more accurate showSourceFile()
 *      nf      08/20/01        Fixed bug 143: don't clear breakpoints
 *      nf      08/21/01        Fix up session state checking;
 *                              stop clearing session properties
 *      nf      08/22/01        Fixed bug 188
 *      nf      08/24/01        Fixed bug 175
 *      nf      08/27/01        Fixed bug 202
 *      tr      12/08/01        Couldn't map source file; more verbose.
 *      nf      12/30/01        Fixed bug 362
 *      MvD     03/10/02        added loadProperties(File) and
 *                              saveProperties(File) and moved the code of
 *                              loadProperties() and saveProperties() to them
 *      nf      03/31/02        Improving the state consistency
 *      nf      04/16/02        Replaced Properties with Preferences
 *      nf      04/29/02        Implemented RFE 507
 *      nf      05/22/02        Print running message at right time
 *      nf      02/07/03        Refactored into interface and class
 *      nf      04/29/03        Fixed bug 727
 *
 * $Id: BasicSession.java 956 2003-04-30 06:55:42Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat;

import com.bluemarsh.jswat.event.SessionEvent;
import com.bluemarsh.jswat.event.SessionListener;
import com.bluemarsh.jswat.event.VMEventListener;
import com.bluemarsh.jswat.event.VMEventManager;
import com.bluemarsh.jswat.ui.UIAdapter;
import com.bluemarsh.jswat.util.ClassUtils;
import com.bluemarsh.jswat.util.StringUtils;
import com.sun.jdi.AbsentInformationException;
import com.sun.jdi.Location;
import com.sun.jdi.ObjectCollectedException;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.VirtualMachine;
import com.sun.jdi.VMDisconnectedException;
import com.sun.jdi.event.Event;
import com.sun.jdi.event.LocatableEvent;
import com.sun.jdi.event.StepEvent;
import com.sun.jdi.event.VMDisconnectEvent;
import com.sun.jdi.event.VMStartEvent;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

/**
 * This class provides a basic implementation of a Session.
 *
 * @author  Nathan Fiedler
 */
public class BasicSession implements Session, VMEventListener {
    /** Session is brand new. */
    private static final int SESSION_NEW = 1;
    /** Session is active. */
    private static final int SESSION_ACTIVE = 2;
    /** Session is inactive. */
    private static final int SESSION_INACTIVE = 3;
    /** Session is presently activating. */
    private static final int SESSION_ACTIVATING = 4;
    /** Session is presently deactivating. */
    private static final int SESSION_DEACTIVATING = 5;
    /** Session is presently closing. */
    private static final int SESSION_CLOSING = 6;
    /** Session is permanently closed. */
    private static final int SESSION_CLOSED = 7;
    /** Logger. */
    private static Logger logger;
    /** Our very own user interface adapter. */
    private UIAdapter interfaceAdapter;
    /** Status log to which messages are written. */
    private Log statusLog;
    /** VMConnection for connecting to the debuggee VM. */
    private VMConnection vmConnection;
    /** Table of Manager objects keyed by their Class. Used to
     * retrieve managers from the Session. */
    private Hashtable managerTable;
    /** List of SessionListener objects. */
    private SessionListenerList listenerList;
    /** Preferences for this class. */
    private Preferences preferences;
    /** State of the Session at this time. */
    private int sessionState;

    static {
        // Initialize the logger.
        logger = Logger.getLogger("com.bluemarsh.jswat.Session");
        com.bluemarsh.jswat.logging.Logging.setInitialState(logger);
    }

    /**
     * Creates a Session object. The session starts off in an inactive
     * state and must be initialized using the <code>init()</code>
     * method.
     *
     * @see #init
     */
    public BasicSession() {
        sessionState = SESSION_NEW;
        // Fetch or create all the objects we need.
        logger.info("creating log, manager table, and listener list");
        statusLog = new Log();
        preferences = Preferences.userRoot().node(
            "com/bluemarsh/jswat/Session");
        managerTable = new Hashtable();
        listenerList = new SessionListenerList();
    } // BasicSession

    /**
     * Activates the Session now that a connection has been made. This
     * method should not be called directly. Instead, the
     * <code>attachDebuggee()</code> or <code>launchDebuggee()</code>
     * methods of <code>VMConnection</code> should be called.
     *
     * @param  connection  VMConnection used to represent the debuggee
     *                     VM connection.
     * @param  showSource  true to show the source for the main class.
     * @param  source      the object invoking this method (will be the
     *                     source of the new session event).
     */
    public synchronized void activate(VMConnection connection,
                                      boolean showSource,
                                      Object source) {
        if (sessionState != SESSION_INACTIVE) {
            throw new IllegalStateException("must be 'inactive'");
        }
        sessionState = SESSION_ACTIVATING;
        boolean success = false;
        SessionEvent se = new SessionEvent(this, source, false);
        try {
            logger.info("activating the session");

            vmConnection = connection;

            // Become a listener to several key VM events.
            // We want to be high priority to get the events
            // before others, to set the context manager.
            VMEventManager vmeman = (VMEventManager)
                getManager(VMEventManager.class);
            vmeman.addListener(VMStartEvent.class, this,
                               VMEventListener.PRIORITY_SESSION);
            vmeman.addListener(VMDisconnectEvent.class, this,
                               VMEventListener.PRIORITY_SESSION);
            vmeman.addListener(StepEvent.class, this,
                               VMEventListener.PRIORITY_SESSION);

            // After we're added as an event listener, we can
            // activate the listeners (including the VMEventManager).
            logger.info("activating session listeners");
            // (we are 'active' as far as the listeners are concerned)
            sessionState = SESSION_ACTIVE;
            listenerList.activateAll(se);
            success = true;
        } finally {
            if (!success) {
                sessionState = SESSION_INACTIVE;
            }
        }
        logger.info("session active");

        if (connection.isRemote()) {
            setStatus(Bundle.getString("Session.vmAttached"));
        } else {
            setStatus(Bundle.getString("Session.vmLoaded"));
        }

        // Let the listeners know if we're running or not.
        if (isSuspended()) {
            listenerList.suspendAll(se);
        } else {
            listenerList.resumeAll(se);
        }

        if (showSource) {
            // Show the source file for the main class.
            String main = connection.getMainClass();
            if (main != null) {
                showSourceFile(main);
            }
        }
    } // activate

    /**
     * Adds a SessionListener to this session. This method
     * calls the <code>init()</code> method on the listener.
     * It may also call the <code>activate()</code> method if
     * the Session is already active.
     *
     * @param  listener  SessionListener to add to this session.
     */
    public void addListener(SessionListener listener) {
        listenerList.add(listener, this, isActive());
        logger.info("adding session listener: "
                    + listener.getClass().getName());
    } // addListener

    /**
     * Shutdown the session permanently. This will close all the
     * listeners, clear all lists and set all references to null.
     *
     * @param  source  the object invoking this method (will be the
     *                 source of the new session event).
     */
    public synchronized void close(Object source) {
        if (sessionState != SESSION_INACTIVE) {
            throw new IllegalStateException("must be 'inactive'");
        }
        sessionState = SESSION_CLOSING;
        boolean success = false;
        try {
            logger.info("session closing");

            // Close all the listeners.
            logger.info("closing session listeners");
            SessionEvent se = new SessionEvent(this, source, false);
            listenerList.closeAll(se);
            // Clear all the lists and references.
            listenerList.clear();
            managerTable.clear();
            success = true;
        } finally {
            if (success) {
                sessionState = SESSION_CLOSED;
            } else {
                sessionState = SESSION_INACTIVE;
            }
        }
        logger.info("session closed");
    } // close

    /**
     * Deactivates this session. If the debuggee VM was lanched, it is
     * terminated. If the debuggee was remote, the session disconnects
     * but leaves the remote debuggee running, unless the 'forceExit'
     * parameter is set to true.
     *
     * @param  forceExit  true to close debuggee VM forcibly in all
     *                    cases; false to leave remote debuggee running.
     * @param  source     the object invoking this method (will be the
     *                    source of the new session event).
     * @see #activate
     */
    public synchronized void deactivate(boolean forceExit, Object source) {
        if (sessionState != SESSION_ACTIVE) {
            throw new IllegalStateException("must be 'active'");
        }
        sessionState = SESSION_DEACTIVATING;
        boolean success = false;
        try {
            logger.info("deactivating the session");

            // Shutdown the running VM before stopping the event handler,
            // so other objects can get the disconnect events.
            VirtualMachine vm = vmConnection.getVM();
            if (vm != null) {
                try {
                    if (vm.process() == null && !forceExit) {
                        // We attached to the VM, so simply detach.
                        logger.info("disposing of debuggee VM");
                        vm.dispose();
                        logger.info("debuggee VM disposed");
                        setStatus(Bundle.getString("Session.vmDetached"));
                    } else {
                        // We launched the VM or the caller wants us to
                        // forcibly kill the debuggee, so bring it down.
                        logger.info("forcing debuggee VM to exit");
                        vm.exit(1);
                        logger.info("debuggee VM terminated");
                        setStatus(Bundle.getString("Session.vmClosed"));
                    }
                } catch (VMDisconnectedException vmde) {
                    // This can happen.
                    logger.info("vm disconnected in deactivate()");
                } finally {
                    vmConnection.disconnect();
                }

            } else {
                if (vmConnection.isRemote()) {
                    setStatus(Bundle.getString("Session.vmDetached"));
                } else {
                    setStatus(Bundle.getString("Session.vmClosed"));
                }
            }

            // Remove ourselves as an event listener.
            VMEventManager vmeman = (VMEventManager)
                getManager(VMEventManager.class);
            vmeman.removeListener(StepEvent.class, this);
            vmeman.removeListener(VMDisconnectEvent.class, this);
            vmeman.removeListener(VMStartEvent.class, this);

            // Notify the listeners that the debugging session has ended.
            logger.info("deactivating session listeners");
            // (we are 'inactive' as far as the listeners are concerned)
            sessionState = SESSION_INACTIVE;
            SessionEvent se = new SessionEvent(this, source, false);
            listenerList.deactivateAll(se);
            success = true;
        } finally {
            if (!success) {
                sessionState = SESSION_ACTIVE;
            }
        }
        logger.info("session inactive");
    } // deactivate

    /**
     * Invoked when a VM event has occurred. The Session object listens
     * for several VM events in order to keep up-to-date with the state
     * of the running debuggee VM.
     *
     * @param  e  VM event
     * @return  true if debuggee VM should be resumed, false otherwise.
     */
    public boolean eventOccurred(Event e) {
        // Note that we will receive VMStartEvents but we do nothing with
        // them. We need to catch them so we can ensure the VM does not
        // resume (which is the default for our event manager).

        if (e instanceof VMDisconnectEvent) {
            // The debuggee VM disconnected; close up shop.
            // When the VM is already disconnected, simply set our
            // reference to null. There's nothing else we can do.
            vmConnection.disconnect();
            setStatus(Bundle.getString("Session.vmDisconnected"));
            if (isActive()) {
                // Call this only if we're still active.
                deactivate(false, this);
            }

        } else if (e instanceof StepEvent) {
            handleDebugEvent((LocatableEvent) e);
        }

        // Most of the time we don't resume the debuggee VM.
        return false;
    } // eventOccurred

    /**
     * Returns the VMConnection object for this Session, if any.
     *
     * @return  VMConnection, or null if none has been made yet.
     */
    public VMConnection getConnection() {
        return vmConnection;
    } // getConnection

    /**
     * Retrieve an instance of a manager of the given class. If one does
     * not exist, an instance will be created.
     *
     * @param  managerClass  Class of Manager to retrieve.
     * @return  Manager object, or null if unlikely things happened.
     */
    public Manager getManager(Class managerClass) {
        Manager manager = (Manager) managerTable.get(managerClass);
        if (manager == null) {
            try {
                // Create the Manager object.
                logger.info("creating manager: " + managerClass.getName());
                manager = (Manager) managerClass.newInstance();
                // Put the Manager in the table.
                managerTable.put(managerClass, manager);
                // Add the Manager as a SessionListener.
                addListener(manager);
            } catch (Exception e) {
                statusLog.writeStackTrace(e);
                // Return the null Manager.
            }
        }
        return manager;
    } // getManager

    /**
     * Searches for the property with the specified key in this property
     * list. The method returns null if the property is not found.
     *
     * @param  key  the property key.
     * @return  the value in this property list with the specified key value.
     */
    public String getProperty(String key) {
        String defaultName = preferences.get("defaultName", "default");
        Preferences prefs = preferences.node(defaultName);
        String value = prefs.get(key, null);
        if (value == null) {
            // See if the value is a long one.
            value = prefs.get(key + ".1", null);
            if (value != null) {
                // It is long, pull all of the pieces together.
                StringBuffer buf = new StringBuffer(
                    Preferences.MAX_VALUE_LENGTH * 2);
                int ki = 1;
                do {
                    buf.append(value);
                    ki++;
                    value = prefs.get(key + '.' + ki, null);
                } while (value != null);
                value = buf.toString();
            }
        }
        return value;
    } // getProperty

    /**
     * Returns the set of session property names.
     *
     * @return  set of session property names, or null if error.
     */
    public String[] getPropertyKeys() {
        try {
            String defaultName = preferences.get("defaultName", "default");
            Preferences prefs = preferences.node(defaultName);
            return prefs.keys();
        } catch (BackingStoreException bse) {
            return null;
        }
    } // getPropertyKeys

    /**
     * Returns a reference to the Log object that receives messages.
     *
     * @return  Log for writing messages to.
     */
    public Log getStatusLog() {
        return statusLog;
    } // getStatusLog

    /**
     * Returns a reference to the virtual machine associated with this
     * session. If the Session is not active, this will return null.
     *
     * @return  Java virtual machine for this session, or null if
     *          not currently connected.
     */
    public VirtualMachine getVM() {
        return vmConnection == null ? null : vmConnection.getVM();
    } // getVM

    /**
     * Returns a reference to the interface adapter associated with
     * this session.
     *
     * @return  interface adapter for this session.
     */
    public UIAdapter getUIAdapter() {
        return interfaceAdapter;
    } // getUIAdapter

    /**
     * Do the usual thing when a debug event has occurred. This involves
     * displaying an indication of the type of event that occurred, such
     * as setting the status indicator. If the event is locatable, this
     * method tries to show the location of the event that occurred, set
     * the current context of the event in the ContextManager, and show
     * the source code for the event location.
     *
     * @param  event  debug event.
     */
    public void handleDebugEvent(Event event) {
        if (!isActive()) {
            // We're not active, there's nothing to do.
            return;
        }

        // Have the interface adapter bring the debugger forward.
        if (preferences.getBoolean("raiseWindow", Defaults.RAISE_WINDOW)) {
            interfaceAdapter.bringForward();
        }

        if (event instanceof LocatableEvent) {
            LocatableEvent le = (LocatableEvent) event;
            // Show where the event occurred.
            showEventLocation(le);

            // Set the debugger context.
            ContextManager contextManager = (ContextManager)
                getManager(ContextManager.class);
            contextManager.setCurrentLocation(le, false);

            // Show the source code for this location.
            showSourceFile(le.location());
        }

        // Treat such events as suspending the debuggee.
        // Must do this after setting the context.
        SessionEvent se = new SessionEvent(this, this, false);
        // We assume here that the event has suspended some threads in
        // the debuggee. If we were unsure, we should check the suspend
        // policy of the event request for this event.
        listenerList.suspendAll(se);
    } // handleDebugEvent

    /**
     * Initialize the Session.
     *
     * @param  uiadapter  interface adapter.
     */
    public synchronized void init(UIAdapter uiadapter) {
        if (sessionState != SESSION_NEW) {
            // We are no longer new, we can't initialize again.
            throw new IllegalStateException("must be 'new'");
        }
        sessionState = SESSION_INACTIVE;
        interfaceAdapter = uiadapter;

        // Some managers need to be loaded at startup.
        String managerList = Bundle.getString("startupManagers");
        String[] managers = StringUtils.tokenize(managerList);
        for (int ii = 0; ii < managers.length; ii++) {
            try {
                getManager(Class.forName(managers[ii]));
            } catch (ClassNotFoundException cnfe) {
                String msg = MessageFormat.format(
                    Bundle.getString("Session.mgrClassNotFound"),
                    new Object[] { managers[ii] });
                getUIAdapter().showMessage(
                    UIAdapter.MESSAGE_WARNING, msg);
            }
        }
    } // init

    /**
     * Returns true if this session is active.
     *
     * @return  true if active, false otherwise.
     * @see #activate
     * @see #deactivate
     */
    public boolean isActive() {
        return sessionState == SESSION_ACTIVE;
    } // isActive

    /**
     * Determines if the debuggee VM has any running threads or not.
     *
     * @return  true if all threads are suspended, false otherwise.
     */
    protected boolean isSuspended() {
        Iterator iter = vmConnection.getVM().allThreads().iterator();
        // Assume they are all suspended.
        boolean allSuspended = true;
        while (iter.hasNext()) {
            ThreadReference thread = (ThreadReference) iter.next();
            try {
                // Checks if the thread was suspended by the debugger.
                if (!thread.isSuspended()) {
                    allSuspended = false;
                    break;
                }
            } catch (ObjectCollectedException oce) {
                // We can ignore that thread then.
            }
        }
        return allSuspended;
    } // isSuspended

    /**
     * Removes a Session listener from this session. This calls the
     * <code>close()</code> method of the listener.
     *
     * @param  listener  SessionListener to remove from this session.
     */
    public void removeListener(SessionListener listener) {
        listenerList.remove(listener, this);
        if (listener instanceof Manager) {
            // Managers must be removed from the manager table.
            logger.info("removing session listener: "
                        + listener.getClass().getName());
            managerTable.remove(listener);
        }
    } // removeListener

    /**
     * Resume execution of the debuggee VM, from a suspended state.
     *
     * @param  source  object making this request.
     * @param  brief   true if caller expects this new state to last
     *                 only a brief amount of time.
     * @param  quiet   true to be silent (do not print anything).
     */
    public void resumeVM(Object source, boolean brief, boolean quiet) {
        if (!isActive()) {
            throw new IllegalStateException("session not active");
        }
        SessionEvent se = new SessionEvent(this, source, brief);
        // Reset both the current thread and location.
        ContextManager contextManager = (ContextManager)
            getManager(ContextManager.class);
        contextManager.setCurrentLocation(null, brief);
        listenerList.resumeAll(se);
        if (!quiet) {
            logger.info("resuming debuggee VM");
            setStatus(Bundle.getString("Session.vmRunning"));
        }
        // Call this after resetting the current location. Otherwise a
        // breakpoint may be hit and the event may fire off
        // asynchronously and we'll then return and clear the current
        // location, just after the breakpoint event had set the current
        // location. Fixes bugs #1 and 2.
        vmConnection.getVM().resume();
    } // resumeVM

    /**
     * Stores the given value in the properties list with the given key
     * as a reference. If the value is null, then the key and value will
     * be removed from the properties.
     *
     * @param  key    the key to be placed into this property list.
     * @param  value  the value corresponding to key, or null to remove
     *                the key and value from the properties.
     * @return  previously stored value, if any.
     */
    public String setProperty(String key, String value) {
        String old = getProperty(key);
        String defaultName = preferences.get("defaultName", "default");
        Preferences prefs = preferences.node(defaultName);
        int ki = 1;
        if (value == null) {
            prefs.remove(key);
        } else {
            int valueLen = value.length();
            if (valueLen > Preferences.MAX_VALUE_LENGTH) {
                // Chop the long value into manageable pieces.
                for (int index = 0; index < valueLen;
                     index += Preferences.MAX_VALUE_LENGTH, ki++) {
                    int end = index + Preferences.MAX_VALUE_LENGTH;
                    if (end > valueLen) {
                        end = valueLen;
                    }
                    prefs.put(key + '.' + ki, value.substring(index, end));
                }
                // Remove the original key, if any, so we are forced
                // to read the long value in getProperty().
                prefs.remove(key);
            } else {
                prefs.put(key, value);
            }
        }
        // Truncate the old long value values.
        String lkey = key + '.' + ki;
        while (prefs.get(lkey, null) != null) {
            prefs.remove(lkey);
            ki++;
            lkey = key + '.' + ki;
        }
        return old;
    } // setProperty

    /**
     * Set the status indicator using a short message. This will call
     * into the interface adapter's <code>showMessage()</code> method.
     *
     * @param  status  short message indicating program status.
     */
    private void setStatus(String status) {
        interfaceAdapter.showMessage(UIAdapter.MESSAGE_NOTICE, status);
    } // setStatus

    /**
     * Shows the location of the event, including the class type, method
     * name, and possibly the source file name and line number (if
     * available).
     *
     * @param  le  LocatableEvent whose location will be displayed.
     */
    protected void showEventLocation(LocatableEvent le) {
        // Show the current thread name.
        StringBuffer buf = new StringBuffer(80);
        buf.append('[');
        buf.append(le.thread().name());
        buf.append("] ");

        Location loc = le.location();
        // Show the class type and method name.
        String tname = loc.declaringType().name();
        tname = ClassUtils.justTheName(tname);
        buf.append(tname);
        buf.append('.');
        buf.append(loc.method().name());
        try {
            // Try to show the source file name and line number.
            String source = loc.sourceName();
            buf.append(" (");
            buf.append(source);
            buf.append(':');
            buf.append(loc.lineNumber());
            buf.append(')');
        } catch (AbsentInformationException aie) {
            // Oh well, no source name then.
        }
        // Do not show this as a notice message.
        statusLog.writeln(buf.toString());
    } // showEventLocation

    /**
     * Show the source file for the given class.
     *
     * @param  loc  location for which to show the source.
     */
    protected void showSourceFile(Location loc) {
        if (!interfaceAdapter.canShowFile()) {
            // Can't show source files, so don't bother.
            return;
        }

        // Try to ensure that the source file is opened.
        PathManager pathman = (PathManager)
            getManager(PathManager.class);
        // This will return the already mapped File, if one exists.
        try {
            SourceSource src = pathman.mapSource(loc.declaringType());
            if (src != null && src.exists()) {
                interfaceAdapter.showFile(src, loc.lineNumber(), 0);
            } else {
                interfaceAdapter.showMessage(
                    UIAdapter.MESSAGE_WARNING,
                    Bundle.getString("Session.couldntMapSrcFile")
                    + " (Method: " + loc.method() + ')');
            }
        } catch (IOException ioe) {
            interfaceAdapter.showMessage(
                UIAdapter.MESSAGE_WARNING,
                Bundle.getString("Session.showFileError") + ": " + ioe);
        }
    } // showSourceFile

    /**
     * Show the source file for the given class.
     *
     * @param  classname  Fully-qualified name of class.
     */
    protected void showSourceFile(String classname) {
        if (!interfaceAdapter.canShowFile()) {
            // Can't show source files, so don't bother.
            return;
        }

        // Try to ensure that the source file is opened.
        PathManager pathman = (PathManager)
            getManager(PathManager.class);
        // This will return the already mapped File, if one exists.
        try {
            SourceSource src = pathman.mapSource(classname);
            if (src != null && src.exists()) {
                interfaceAdapter.showFile(src, 0, 0);
            } else {
                interfaceAdapter.showMessage(
                    UIAdapter.MESSAGE_WARNING,
                    Bundle.getString("Session.couldntMapSrcFile")
                    + " (Class: " + classname + " )");
            }
        } catch (IOException ioe) {
            interfaceAdapter.showMessage(
                UIAdapter.MESSAGE_WARNING,
                Bundle.getString("Session.showFileError") + ": " + ioe);
        }
    } // showSourceFile

    /**
     * Suspend execution of the debuggee VM.
     *
     * @param  source  object making this request.
     */
    public void suspendVM(Object source) {
        if (!isActive()) {
            throw new IllegalStateException("session not active");
        }
        // See if there are any running threads.
        if (!isSuspended()) {
            // There are still running threads, let us suspend them.
            vmConnection.getVM().suspend();
            logger.info("suspending debuggee VM");
            setStatus(Bundle.getString("Session.vmSuspended"));
            SessionEvent se = new SessionEvent(this, source, false);
            listenerList.suspendAll(se);
        }
    } // suspendVM
} // BasicSession
