/*********************************************************************
 *
 *      Copyright (C) 2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      JSwat
 * FILE:        ByteCodeSource.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      03/04/03        Initial version
 *      nf      04/07/03        Fixed bug 749
 *
 * $Id: ByteCodeSource.java 925 2003-04-08 05:06:56Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat;

import com.sun.jdi.ReferenceType;
import java.io.InputStream;

/**
 * ByteCodeSource provides a means of reading byte code from a class in
 * debuggee VM.
 *
 * @author  Nathan Fiedler
 */
public class ByteCodeSource implements SourceSource {
    /** The class we are reading bytecode from. */
    private ReferenceType clazz;

    /**
     * Constructs a ByteCodeSource instance for the given class.
     *
     * @param  clazz  class to read bytecode from.
     */
    public ByteCodeSource(ReferenceType clazz) {
        this.clazz = clazz;
    } // ByteCodeSource

    /**
     * Indicates whether some other object is "equal to" this one.
     *
     * @param  o  the reference object with which to compare.
     * @return  true if this object is the same as the obj argument;
     *          false otherwise.
     */
    public boolean equals(Object o) {
        if (o instanceof ByteCodeSource) {
            ByteCodeSource obcs = (ByteCodeSource) o;
            return obcs.clazz.equals(clazz);
        }
        return false;
    } // equals

    /**
     * Attempts to ensure that the referenced source actually exists.
     *
     * @return  true if source exists, false if not found.
     */
    public boolean exists() {
        return true;
    } // exists

    /**
     * Returns just the name of the source file, not including the path
     * to the file, if any.
     *
     * @return  name of source.
     */
    public String getName() {
        return clazz.name();
    } // getName

    /**
     * Get the input stream for reading the source code.
     *
     * @return  input stream to the source code.
     */
    public InputStream getInputStream() {
        return null;
    } // getInputStream

    /**
     * Returns the class this source is to read from.
     *
     * @return  class type.
     */
    public ReferenceType getReferenceType() {
        return clazz;
    } // getReferenceType

    /**
     * Returns a hash code value for the object.
     *
     * @return  a hash code value for this object.
     */
    public int hashCode() {
        // Hopefully the file is resolved to the canonical form
        // before the hashcode is calculated.
        return clazz.hashCode();
    } // hashCode
} // ByteCodeSource
