/*********************************************************************
 *
 *      Copyright (C) 2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Expression
 * FILE:        DivBinaryOperatorNode.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      06/10/02        Initial version
 *
 * $Id: DivBinaryOperatorNode.java 662 2002-11-02 07:25:59Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.expr;

import com.bluemarsh.jswat.parser.java.node.Token;

/**
 * Class DivBinaryOperatorNode implements the division binary operator (/).
 *
 * @author  Nathan Fiedler
 */
class DivBinaryOperatorNode extends BinaryOperatorNode {

    /**
     * Constructs a DivBinaryOperatorNode.
     *
     * @param  node  lexical token.
     */
    public DivBinaryOperatorNode(Token node) {
        super(node);
    } // DivBinaryOperatorNode

    /**
     * Returns the value of this node.
     *
     * @param  context  evaluation context.
     * @return  a Number.
     * @throws  EvaluationException
     *          if an error occurred during evaluation.
     */
    public Object eval(EvaluationContext context) throws EvaluationException {
        Node n1 = getChild(0);
        Object o1 = n1.eval(context);
        Node n2 = getChild(1);
        Object o2 = n2.eval(context);
        if (isNumber(o1) && isNumber(o2)) {
            if (!isFloating(o1) && !isFloating(o2)) {
                // Both numbers are integer values.
                long l1 = getLongValue(o1);
                long l2 = getLongValue(o2);
                if (l2 == 0L) {
                    throw new EvaluationException(
                        Bundle.getString("div.byZero"), getToken());
                }
                return makeSmallestInteger(l1 / l2);
            } else {
                // One or both are floating point values.
                double d1 = getDoubleValue(o1);
                double d2 = getDoubleValue(o2);
                if (d2 == 0.0) {
                    throw new EvaluationException(
                        Bundle.getString("div.byZero"), getToken());
                }
                return makeSmallestFloat(d1 / d2);
            }
        } else {
            throw new EvaluationException(
                Bundle.getString("errorNumericOperator"), getToken());
        }
    } // eval

    /**
     * Returns this operator's precedence value. The lower the value
     * the higher the precedence. The values are equivalent to those
     * described in the Java Language Reference book (2nd ed.), p 106.
     *
     * @return  precedence value.
     */
    public int precedence() {
        return 6;
    } // precedence
} // DivBinaryOperatorNode
