/*********************************************************************
 *
 *      Copyright (C) 2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Expression
 * FILE:        EqualsOperatorNode.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      06/11/02        Initial version
 *
 * $Id: EqualsOperatorNode.java 662 2002-11-02 07:25:59Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.expr;

import com.bluemarsh.jswat.parser.java.node.Token;

/**
 * Class EqualsOperatorNode implements the equals to operator (==).
 *
 * @author  Nathan Fiedler
 */
class EqualsOperatorNode extends BinaryOperatorNode {

    /**
     * Constructs a EqualsOperatorNode.
     *
     * @param  node  lexical token.
     */
    public EqualsOperatorNode(Token node) {
        super(node);
    } // EqualsOperatorNode

    /**
     * Returns the value of this node.
     *
     * @param  context  evaluation context.
     * @return  a Boolean.
     * @throws  EvaluationException
     *          if an error occurred during evaluation.
     */
    public Object eval(EvaluationContext context) throws EvaluationException {
        Node n1 = getChild(0);
        Object o1 = n1.eval(context);
        Node n2 = getChild(1);
        Object o2 = n2.eval(context);

        if (o1 == null || o2 == null) {
            return o1 == o2 ? Boolean.TRUE : Boolean.FALSE;

        } else if (isBoolean(o1) || isBoolean(o2)) {
            if (!isBoolean(o1) || !isBoolean(o2)) {
                throw new EvaluationException(
                    Bundle.getString("equal.sameType"), getToken());
            } else {
                boolean b1 = getBooleanValue(o1);
                boolean b2 = getBooleanValue(o2);
                return b1 == b2 ? Boolean.TRUE : Boolean.FALSE;
            }

        } else if (isNumber(o1) || isNumber(o2)) {
            if (!isNumber(o1) || !isNumber(o2)) {
                throw new EvaluationException(
                    Bundle.getString("equal.sameType"), getToken());
            } else {
                if (!isFloating(o1) && !isFloating(o2)) {
                    long l1 = getLongValue(o1);
                    long l2 = getLongValue(o2);
                    return l1 == l2 ? Boolean.TRUE : Boolean.FALSE;
                } else {
                    double d1 = getDoubleValue(o1);
                    double d2 = getDoubleValue(o2);
                    return d1 == d2 ? Boolean.TRUE : Boolean.FALSE;
                }
            }

        } else if (isCharacter(o1) || isCharacter(o2)) {
            if (!isCharacter(o1) || !isCharacter(o2)) {
                throw new EvaluationException(
                    Bundle.getString("equal.sameType"), getToken());
            } else {
                char c1 = getCharValue(o1);
                char c2 = getCharValue(o2);
                return c1 == c2 ? Boolean.TRUE : Boolean.FALSE;
            }

        } else {
            return o1.equals(o2) ? Boolean.TRUE : Boolean.FALSE;
        }
    } // eval

    /**
     * Returns this operator's precedence value. The lower the value
     * the higher the precedence. The values are equivalent to those
     * described in the Java Language Reference book (2nd ed.), p 106.
     *
     * @return  precedence value.
     */
    public int precedence() {
        return 10;
    } // precedence
} // EqualsOperatorNode
