/*********************************************************************
 *
 *      Copyright (C) 2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Expression
 * FILE:        MethodNode.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      06/04/02        Initial version
 *      nf      12/28/02        Fixed bug 690
 *
 * $Id: MethodNode.java 962 2003-05-22 04:55:52Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.expr;

import com.bluemarsh.jswat.Defaults;
import com.bluemarsh.jswat.breakpoint.AmbiguousClassSpecException;
import com.bluemarsh.jswat.breakpoint.AmbiguousMethodException;
import com.bluemarsh.jswat.parser.java.node.Token;
import com.bluemarsh.jswat.util.ClassUtils;
import com.bluemarsh.jswat.util.VariableUtils;
import com.sun.jdi.*;
import java.util.*;
import java.util.prefs.Preferences;

/**
 * Class MethodNode represents a method invocation.
 *
 * @author  Nathan Fiedler
 */
class MethodNode extends OperatorNode {

    /**
     * Constructs a MethodNode associated with the given token.
     *
     * @param  node  lexical token.
     */
    public MethodNode(Token node) {
        super(node);
    } // MethodNode

    /**
     * Returns a list of Strings representing the types of the
     * arguments in the list given.
     *
     * @param  arguments  list of arguments.
     * @return  list of argument types.
     */
    protected static List argumentTypes(List arguments) {
        List types = new ArrayList(arguments.size());

        for (int ii = 0; ii < arguments.size(); ii++) {
            Object o = arguments.get(ii);
            if (o instanceof String) {
                types.add("java.lang.String");
            } else if (o instanceof Boolean) {
                types.add("java.lang.Boolean");
            } else if (o instanceof Character) {
                types.add("java.lang.Character");
            } else if (o instanceof Double) {
                types.add("java.lang.Double");
            } else if (o instanceof Float) {
                types.add("java.lang.Float");
            } else if (o instanceof Integer) {
                types.add("java.lang.Integer");
            } else if (o instanceof Long) {
                types.add("java.lang.Long");
            } else if (o instanceof Short) {
                types.add("java.lang.Short");
            } else if (o instanceof Byte) {
                types.add("java.lang.Byte");
            } else if (o instanceof Value) {
                Value v = (Value) o;
                types.add(v.type().name());
            } else {
                types.add("java.lang.Object");
            }
        }

        return types;
    } // argumentTypes

    /**
     * Returns the value of this node.
     *
     * @param  context  evaluation context.
     * @return  value.
     * @throws  EvaluationException
     *          if an error occurred during evaluation.
     */
    public Object eval(EvaluationContext context) throws EvaluationException {
        // Get the preliminaries out of the way first.
        ThreadReference thread = context.getThread();
        if (thread == null) {
            throw new EvaluationException(
                Bundle.getString("method.noThread"));
        }
        int frameIdx = context.getFrame();
        VirtualMachine vm = thread.virtualMachine();

        // First argument had better be the method name.
        IdentifierNode idn = (IdentifierNode) getChild(0);
        String methodName = idn.getName();

        // XXX - Change the method node to use JNI-style signatures for
        // finding methods. This will make matching the types a bit
        // easier. And it removes the need to be able to load the
        // referenced classes in the JSwat VM. If an argument type does
        // not match exactly, get the ClassType from JDI and check for
        // matches with superclasses and implemented interfaces.

        // Get the object or class referenced by the first part
        // of the method identifier. If no first part given,
        // infer from the current location in the debuggee.
        ObjectReference objref = null;
        ReferenceType reftype = null;
        int lastDot = methodName.lastIndexOf('.');
        if (lastDot == -1) {
            // First part was not given, default to class containing
            // current location.
            try {
                if (thread.frameCount() == 0) {
                    throw new EvaluationException(
                        Bundle.getString("method.threadNotRunning"));
                }
                StackFrame frame = thread.frame(frameIdx);
                objref = frame.thisObject();

                if (objref == null) {
                    Location location = context.getLocation();
                    if (location == null) {
                        throw new EvaluationException(
                            Bundle.getString("method.unknownLocation"));
                    }
                    reftype = location.declaringType();
                    if (reftype == null) {
                        throw new EvaluationException(
                            Bundle.getString("method.unknownLocation"));
                    }
                } else {
                    reftype = objref.referenceType();
                }
            } catch (IncompatibleThreadStateException itse) {
                throw new EvaluationException(
                    Bundle.getString("method.badThreadState"));
            }

        } else {
            // Is the first part a variable or a class name?
            String varOrClass = methodName.substring(0, lastDot);
            try {
                Value variable = VariableUtils.getValue(varOrClass, thread,
                                                        frameIdx);
                objref = (ObjectReference) variable;
                reftype = objref.referenceType();
            } catch (Exception e) {
                // Maybe it's not a variable reference.
            }

            if (reftype == null) {
                // See if the argument is a class name.
                List classes = ClassUtils.findClassesByPattern(
                    vm, varOrClass);
                if (classes == null || classes.size() == 0) {
                    throw new EvaluationException(
                        Bundle.getString("method.badVarOrClass"));
                }
                reftype = (ReferenceType) classes.get(0);
            }
        }

        // Get the list of method arguments.
        List methodArgs = new ArrayList();
        int count = childCount();
        for (int ii = 1; ii < count; ii++) {
            Node n = getChild(ii);
            Object o = n.eval(context);
            methodArgs.add(o);
        }

        // Make a list of argument types, based on the types of the
        // arguments passed from the user.
        List argTypes = argumentTypes(methodArgs);

        // Locate method in the resolved class. The ClassUtils method
        // will employ fuzzy logic to find a reasonable match. This
        // means it will treat primitives and their equivalent class
        // types as being the same; in addition, it considers the
        // possibility of up-casting numeric types.
        Method method = null;
        try {
            String methodId = methodName.substring(lastDot + 1);
            method = ClassUtils.findMethod(reftype, methodId, argTypes);
        } catch (AmbiguousClassSpecException acse) {
            throw new EvaluationException(
                Bundle.getString("method.ambiguousClass"));
        } catch (AmbiguousMethodException acse) {
            throw new EvaluationException(
                Bundle.getString("method.ambiguousMethod"));
        } catch (NoSuchMethodException nsme) {
            throw new EvaluationException(
                Bundle.getString("method.noSuchMethod"));
        }

        // Convert the arguments to JDI objects.
        mirrorArguments(vm, methodArgs);

        // Invoke the method and return the results.
        Preferences prefs = Preferences.userRoot().node(
            "com/bluemarsh/jswat/util");
        int timeout = prefs.getInt("invocationTimeout",
                                   Defaults.INVOCATION_TIMEOUT);
        try {
            return ClassUtils.invokeMethod(objref, reftype, thread, method,
                                           methodArgs, timeout);
        } catch (Exception e) {
            throw new EvaluationException(e);
        }
    } // eval

    /**
     * Translate the given list of arguments into Value instances of
     * the appropriate type. The list is modified in-place.
     *
     * @param  vm    debuggee virtual machine.
     * @param  args  list of arguments to mirror.
     */
    protected static void mirrorArguments(VirtualMachine vm, List args) {
        for (int ii = 0; ii < args.size(); ii++) {
            Object o = args.get(ii);
            if (o instanceof String) {
                args.set(ii, vm.mirrorOf((String) o));
            } else if (o instanceof Boolean) {
                args.set(ii, vm.mirrorOf(((Boolean) o).booleanValue()));
            } else if (o instanceof Character) {
                args.set(ii, vm.mirrorOf(((Character) o).charValue()));
            } else if (o instanceof Double) {
                args.set(ii, vm.mirrorOf(((Double) o).doubleValue()));
            } else if (o instanceof Float) {
                args.set(ii, vm.mirrorOf(((Float) o).floatValue()));
            } else if (o instanceof Integer) {
                args.set(ii, vm.mirrorOf(((Integer) o).intValue()));
            } else if (o instanceof Long) {
                args.set(ii, vm.mirrorOf(((Long) o).longValue()));
            } else if (o instanceof Short) {
                args.set(ii, vm.mirrorOf(((Short) o).shortValue()));
            } else if (o instanceof Byte) {
                args.set(ii, vm.mirrorOf(((Byte) o).byteValue()));
            }
            // Else it's a null or Value.
        }
    } // mirrorArguments

    /**
     * Returns this operator's precedence value. The lower the value
     * the higher the precedence. The values are equivalent to those
     * described in the Java Language Reference book (2nd ed.), p 106.
     *
     * @return  precedence value.
     */
    public int precedence() {
        return 1;
    } // precedence
} // MethodNode
