/*********************************************************************
 *
 *      Copyright (C) 2002-2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Expression
 * FILE:        OperatorNode.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      06/02/02        Initial version
 *
 * $Id: OperatorNode.java 826 2003-01-29 08:07:47Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.expr;

import com.bluemarsh.jswat.parser.java.node.Token;
import com.sun.jdi.BooleanValue;
import com.sun.jdi.ByteValue;
import com.sun.jdi.CharValue;
import com.sun.jdi.DoubleValue;
import com.sun.jdi.FloatValue;
import com.sun.jdi.IntegerValue;
import com.sun.jdi.LongValue;
import com.sun.jdi.PrimitiveValue;
import com.sun.jdi.ShortValue;

/**
 * Class OperatorNode is the base class for all operators.
 *
 * @author  Nathan Fiedler
 */
public abstract class OperatorNode extends ParentNode {

    /**
     * Determines the type of 'o' and returns true if it is a boolean.
     *
     * @param  o  object to test.
     * @return  true if 'o' is a boolean.
     */
    public static boolean isBoolean(Object o) {
        return (o instanceof Boolean || o instanceof BooleanValue);
    } // isBoolean

    /**
     * Determines the type of 'o' and returns true if it is a character.
     *
     * @param  o  object to test.
     * @return  true if 'o' is a character.
     */
    public static boolean isCharacter(Object o) {
        return (o instanceof Character || o instanceof CharValue);
    } // isCharacter

    /**
     * Determines the type of 'o' and returns true if it is a
     * floating point number.
     *
     * @param  o  object to test.
     * @return  true if 'o' is a number.
     */
    public static boolean isFloating(Object o) {
        return (o instanceof Double
                || o instanceof Float
                || o instanceof DoubleValue
                || o instanceof FloatValue);
    } // isFloating

    /**
     * Determines the type of 'o' and returns true if it is some
     * type of number, either integral or floating.
     *
     * @param  o  object to test.
     * @return  true if 'o' is a number.
     */
    public static boolean isNumber(Object o) {
        return (o instanceof Number
                || o instanceof ByteValue
                || o instanceof DoubleValue
                || o instanceof FloatValue
                || o instanceof IntegerValue
                || o instanceof LongValue
                || o instanceof ShortValue);
    } // isNumber

    /**
     * Return the best-fitting Number type for the given value.
     *
     * @param  d  value to wrap with a Number instance.
     * @return  best-fitting Number wrapper.
     */
    public static Number makeSmallestFloat(double d) {
        if (d <= Float.MAX_VALUE) {
            return new Float((float) d);
        } else {
            return new Double(d);
        }
    } // makeSmallestFloat

    /**
     * Return the best-fitting Number type for the given value.
     *
     * @param  l  value to wrap with a Number instance.
     * @return  best-fitting Number wrapper.
     */
    public static Number makeSmallestInteger(long l) {
        if (l <= Byte.MAX_VALUE) {
            return new Byte((byte) l);
        } else if (l <= Short.MAX_VALUE) {
            return new Short((short) l);
        } else if (l <= Integer.MAX_VALUE) {
            return new Integer((int) l);
        } else {
            return new Long(l);
        }
    } // makeSmallestInteger

    /**
     * Constructs a OperatorNode associated with the given token.
     *
     * @param  node  lexical token.
     */
    public OperatorNode(Token node) {
        super(node);
    } // OperatorNode

    /**
     * Returns the boolean value from the given object.
     *
     * @param  o  object that had better be a Boolean type.
     * @return  boolean value of object.
     * @throws  EvaluationException
     *          if o is not a boolean type.
     */
    public boolean getBooleanValue(Object o) throws EvaluationException {
        if (o instanceof Boolean) {
            Boolean b = (Boolean) o;
            return b.booleanValue();
        } else if (o instanceof BooleanValue) {
            BooleanValue bv = (BooleanValue) o;
            return bv.value();
        } else {
            throw new EvaluationException(Bundle.getString("errorNotBoolean"),
                                          getToken());
        }
    } // getBooleanValue

    /**
     * Returns the char value from the given object.
     *
     * @param  o  object that had better be a Character type.
     * @return  char value of object.
     * @throws  EvaluationException
     *          if o is not a character type.
     */
    public char getCharValue(Object o) throws EvaluationException {
        if (o instanceof Character) {
            Character b = (Character) o;
            return b.charValue();
        } else if (o instanceof CharValue) {
            CharValue cv = (CharValue) o;
            return cv.value();
        } else {
            throw new EvaluationException(Bundle.getString("errorNotChar"),
                                          getToken());
        }
    } // getCharValue

    /**
     * Returns the double value from the given object.
     *
     * @param  o  object that had better be a Double type.
     * @return  double value of object.
     * @throws  EvaluationException
     *          if o is not a double type.
     */
    public double getDoubleValue(Object o) throws EvaluationException {
        if (o instanceof Number) {
            Number n = (Number) o;
            return n.doubleValue();
        } else if (o instanceof PrimitiveValue) {
            PrimitiveValue pv = (PrimitiveValue) o;
            return pv.doubleValue();
        } else {
            throw new EvaluationException(Bundle.getString("errorNotDouble"),
                                          getToken());
        }
    } // getDoubleValue

    /**
     * Returns the long value from the given object.
     *
     * @param  o  object that had better be a Long type.
     * @return  long value of object.
     * @throws  EvaluationException
     *          if o is not a long type.
     */
    public long getLongValue(Object o) throws EvaluationException {
        if (o instanceof Number) {
            Number n = (Number) o;
            return n.longValue();
        } else if (o instanceof PrimitiveValue) {
            PrimitiveValue pv = (PrimitiveValue) o;
            return pv.longValue();
        } else {
            throw new EvaluationException(Bundle.getString("errorNotLong"),
                                          getToken());
        }
    } // getLongValue

    /**
     * Returns true if this operator does not do any operation but
     * instead acts as a sentinel, delineating portions of an
     * expression. This includes (), [], and commas.
     *
     * @return  true if sentinel, false otherwise.
     */
    public boolean isSentinel() {
        return false;
    } // isSentinel

    /**
     * Returns this operator's precedence value. The lower the value
     * the higher the precedence. The values are equivalent to those
     * described in the Java Language Reference book (2nd ed.), p 106.
     *
     * @return  precedence value.
     */
    public abstract int precedence();
} // OperatorNode
