/*********************************************************************
 *
 *      Copyright (C) 2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Expression
 * FILE:        Types.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      05/19/03        Initial version
 *
 * $Id: Types.java 0 2003-05-22 04:55:52Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.expr;

/**
 * Class Types provides utility methods for handling types of various
 * forms, especially converting to and from JNI style type signatures.
 *
 * @author  Nathan Fiedler
 */
public class Types {

    /**
     * We are not instantiated.
     */
    private Types() {
    } // Types

    /**
     * Converts a type name into a JNI type signature.
     *
     * @param  type  type name (e.g. "boolean", "java.net.URL", "String").
     * @return  JNI type signature (e.g. "Z", "Ljava/net/URL;",
     *          "Ljava/lang/String;"), or null if not recognized.
     */
    public static String typeNameToJNI(String type) {
        String sig = null;
        // Check if it is a primitive type.
        if (type.equals("boolean")) {
            sig = "Z";
        } else if (type.equals("byte")) {
            sig = "B";
        } else if (type.equals("char")) {
            sig = "C";
        } else if (type.equals("double")) {
            sig = "D";
        } else if (type.equals("float")) {
            sig = "F";
        } else if (type.equals("int")) {
            sig = "I";
        } else if (type.equals("long")) {
            sig = "J";
        } else if (type.equals("short")) {
            sig = "S";
        } else if (type.equals("void")) {
            sig = "V";

        } else {
            try {
                // See if it is a core class.
                Class clazz = Class.forName("java.lang." + type);
                sig = "Ljava/lang/" + type + ';';
            } catch (ClassNotFoundException cnfe) {
                // Must be some other class.
                sig = 'L' + type.replace('.', '/') + ';';
            }
        }

        // Check if the type is an array type.
        if (type.endsWith("[]")) {
            return '[' + sig;
        } else {
            return sig;
        }
    } // typeNameToJNI
} // Types
