/*********************************************************************
 *
 *      Copyright (C) 2001-2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      JSwat UI
 * FILE:        BasicUIAdapter.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      09/11/01        Initial version
 *      nf      12/27/02        Added property change support
 *
 * $Id: BasicUIAdapter.java 879 2003-03-09 07:18:42Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.ui;

import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.SourceSource;
import com.bluemarsh.jswat.view.View;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.net.URL;
import java.util.Hashtable;

/**
 * Abstract class BasicUIAdapter provides a limited implementation of
 * the UIAdapter interface. Concrete adapter implementations should
 * extend this class.
 *
 * @author  Nathan Fiedler
 */
public abstract class BasicUIAdapter implements UIAdapter {
    /** Table of properties. */
    protected Hashtable propertyTable;
    /** Used for signaling changes in bound properties. */
    protected PropertyChangeSupport changeSupport;

    /**
     * Constructor for BasicUIAdapter class.
     */
    public BasicUIAdapter() {
        propertyTable = new Hashtable();
        changeSupport = new PropertyChangeSupport(this);
    } // BasicUIAdapter

    /**
     * Add a PropertyChangeListener to the listener list. The listener is
     * registered for all properties.
     *
     * @param  listener  the PropertyChangeListener to be added.
     */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.addPropertyChangeListener(listener);
    } // addPropertyChangeListener

    /**
     * Add a PropertyChangeListener for a specific property. The listener
     * will be invoked only when a call on firePropertyChange names that
     * specific property.
     *
     * @param  propertyName  the name of the property to listen on.
     * @param  listener      the PropertyChangeListener to be added.
     */

    public void addPropertyChangeListener(String propertyName,
                                          PropertyChangeListener listener) {
        changeSupport.addPropertyChangeListener(propertyName, listener);
    } // addPropertyChangeListener

    /**
     * In a graphical environment, bring the primary debugger window
     * forward so the user can see it. This is called primarily when a
     * debugger event has occurred and the debugger may be hidden behind
     * the debuggee application window.
     */
    public abstract void bringForward();

    /**
     * Construct the appropriate user interface and connect all the
     * pieces together. The result should be a fully functional
     * interface that is ready to be used.
     */
    public abstract void buildInterface();

    /**
     * Indicate if this interface adapter has the ability to find a
     * string in the currently selected source view.
     *
     * @return  true if the ability exists, false otherwise.
     */
    public abstract boolean canFindString();

    /**
     * Indicate if this interface adapter has the ability to show source
     * files in a manner appropriate for the user to read.
     *
     * @return  true if the ability exists, false otherwise.
     */
    public abstract boolean canShowFile();

    /**
     * Deconstruct the user interface such that all components are made
     * invisible and prepared for non-use.
     */
    public abstract void destroyInterface();

    /**
     * This is called when there are no more open Sessions. The adapter
     * should take the appropriate action at this time. In most cases
     * that will be to exit the JVM.
     */
    public abstract void exit();

    /**
     * Search for the given string in the currently selected source
     * view. The search should continue from the last successful match,
     * and wrap around to the beginning when the end is reached.
     *
     * @param  query       string to look for.
     * @param  ignoreCase  true to ignore case.
     * @return  true if string was found.
     * @exception  NoOpenViewException
     *             if there is no view to be searched.
     */
    public abstract boolean findString(String query, boolean ignoreCase)
        throws NoOpenViewException;

    /**
     * Searches for the property with the specified key in the property
     * list. The method returns null if the property is not found.
     *
     * @param  key  the property key.
     * @return  the value in the property list with the specified key value.
     */
    public Object getProperty(String key) {
        return propertyTable.get(key);
    } // getProperty

    /**
     * Retrieves the currently active view in JSwat.
     *
     * @return  selected view, or null if none selected.
     */
    public abstract View getSelectedView();

    /**
     * Perform any initialization that requires a Session instance. This is
     * called after the object is constructed and before
     * <code>buildInterface()</code> is called.
     *
     * @param  session  session to associate with.
     */
    public abstract void init(Session session);

    /**
     * Called when the Session initialization has completed.
     */
    public abstract void initComplete();

    /**
     * Refresh the display to reflect changes in the program. Generally
     * this means refreshing the panels.
     */
    public abstract void refreshDisplay();

    /**
     * Remove a PropertyChangeListener from the listener list. This removes
     * a PropertyChangeListener that was registered for all properties.
     *
     * @param  listener  the PropertyChangeListener to be removed.
     */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.removePropertyChangeListener(listener);
    } // removePropertyChangeListener

    /**
     * Remove a PropertyChangeListener for a specific property.
     *
     * @param  propertyName  the name of the property that was listened on.
     * @param  listener      the PropertyChangeListener to be removed.
     */

    public void removePropertyChangeListener(String propertyName,
                                             PropertyChangeListener listener) {
        changeSupport.removePropertyChangeListener(propertyName, listener);
    } // removePropertyChangeListener

    /**
     * Save any settings to the appropriate places, the program is about
     * to terminate.
     */
    public abstract void saveSettings();

    /**
     * Stores the given value in the properties list with the given key
     * as a reference. If the value is null, then the key and value will
     * be removed from the properties.
     *
     * @param  key    the key to be placed into this property list.
     * @param  value  the value corresponding to key, or null to remove
     *                the key and value from the properties.
     * @return  previous value stored using this key.
     */
    public Object setProperty(String key, Object value) {
        if (value == null) {
            return propertyTable.remove(key);
        } else {
            return propertyTable.put(key, value);
        }
    } // setProperty

    /**
     * Show the given file in the appropriate view and make the given
     * line visible in that view.
     *
     * @param  src    source to be displayed.
     * @param  line   one-based line to be made visible, or zero for
     *                a reasonable default.
     * @param  count  number of lines to display, or zero for a
     *                reasonable default. Some adapters will ignore
     *                this value if, for instance, they utilize a
     *                scrollable view.
     * @return  true if successful, false if error.
     */
    public abstract boolean showFile(SourceSource src, int line, int count);

    /**
     * Show a help screen written in HTML. This is may be implemented
     * like the <code>showURL()</code> method, but should have buttons
     * for navigating the help content.
     *
     * @param  url  help screen to be shown to the user.
     */
    public abstract void showHelp(URL url);

    /**
     * Show a message in an appropriate location.
     *
     * @param  type  one of the message types defined in this class.
     * @param  msg   message to be shown to the user.
     */
    public abstract void showMessage(int type, String msg);

    /**
     * Show a URL in a reasonable manner. This will likely involve using
     * a <code>JEditorPane</code> or some similar class to display the
     * file referenced by the <code>URL</code>.
     *
     * @param  url    URL to be shown to the user.
     * @param  title  title for the window showing the URL, if any.
     */
    public abstract void showURL(URL url, String title);
} // BasicUIAdapter
