/*********************************************************************
 *
 *      Copyright (C) 2002-2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      JSwat UI
 * FILE:        OpenedFilesMenu.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      12/27/02        Initial version
 *
 * $Id: OpenedFilesMenu.java 832 2003-01-30 07:06:42Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.ui.graphical;

import com.bluemarsh.jswat.Defaults;
import com.bluemarsh.jswat.FileSource;
import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.ui.Bundle;
import com.bluemarsh.jswat.ui.SessionFrameMapper;
import com.bluemarsh.jswat.ui.UIAdapter;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.prefs.Preferences;
import javax.swing.JMenu;
import javax.swing.JMenuItem;

/**
 * This is a specialized menu class that implements the most recently
 * opened files menu. This menu will hold a list of file names representing
 * the files that were recently opened, up to a fixed number of files.
 * Selecting a menu item will reopen that file in a source view.
 *
 * @author  Nathan Fiedler
 */
class OpenedFilesMenu extends JMenu
    implements ActionListener, PropertyChangeListener {
    /** Maximum length in characters for menu items. */
    private static final int MENU_MAX_LENGTH = 50;
    /** Prefix for the file node names. */
    private static final String NODE_PREFIX = "file";
    /** List of recent files. */
    private List recentFiles;
    /** Our preferences node. */
    private Preferences preferences;

    /**
     * Constructor for this class.
     *
     * @param  adapter  interface adapter we belong to.
     */
    public OpenedFilesMenu(UIAdapter adapter) {
        super(Bundle.getString("openedFilesLabel"), true);
        recentFiles = new LinkedList();
        adapter.addPropertyChangeListener(this);
        setToolTipText("<html><small>"
                       + Bundle.getString("openedFilesTooltip")
                       + "</small></html>");
        preferences = Preferences.userRoot().node(
            "com/bluemarsh/jswat/ui/mro");

        // Populate the list from the preferences.
        int maxSize = preferences.getInt("maxListSize",
                                         Defaults.MRO_LIST_SIZE);
        for (int ii = maxSize; ii >= 1; ii--) {
            String val = preferences.get(NODE_PREFIX + ii, null);
            if (val != null) {
                JMenuItem item = createMenuItem(val);
                FileSource fs = new FileSource(new File(val));
                FileItemPair fip = new FileItemPair(fs, item);
                recentFiles.add(fip);
            }
        }
    } // OpenedFilesMenu

    /**
     * A menu item was selected.
     *
     * @param  evt  the action event.
     */
    public void actionPerformed(ActionEvent evt) {
        JMenuItem src = (JMenuItem) evt.getSource();
        // Find the menu-file pair in the list.
        FileItemPair fip = findPair(src);
        if (fip != null) {
            Session session = SessionFrameMapper.getSessionForEvent(evt);
            UIAdapter adapter = session.getUIAdapter();
            adapter.showFile(fip.getFileSource(), 0, 0);
        }
    } // actionPerformed

    /**
     * Builds a menu item for file name.
     *
     * @param  filename  full path of the file.
     * @return  new menu item.
     */
    protected JMenuItem createMenuItem(String filename) {
        int len = filename.length();
        if (len > MENU_MAX_LENGTH) {
            filename = "..." + filename.substring(len - MENU_MAX_LENGTH);
        }
        JMenuItem item = new JMenuItem(filename);
        item.addActionListener(this);
        // Add to the top of the list.
        add(item, 0);
        return item;
    } // createMenuItem

    /**
     * Find the FileItemPair by looking for the given file source.
     *
     * @param  src  file source to look for.
     * @return  FileItemPair if found, null if not.
     */
    protected FileItemPair findPair(FileSource src) {
        for (int ii = recentFiles.size() - 1; ii >= 0; ii--) {
            FileItemPair fip = (FileItemPair) recentFiles.get(ii);
            if (fip.getFileSource().equals(src)) {
                return fip;
            }
        }
        return null;
    } // findPair

    /**
     * Find the FileItemPair by looking for the given menu item.
     *
     * @param  item  menu item to look for.
     * @return  FileItemPair if found, null if not.
     */
    protected FileItemPair findPair(JMenuItem item) {
        for (int ii = recentFiles.size() - 1; ii >= 0; ii--) {
            FileItemPair fip = (FileItemPair) recentFiles.get(ii);
            if (fip.getMenuItem().equals(item)) {
                return fip;
            }
        }
        return null;
    } // findPair

    /**
     * This method gets called when a bound property is changed.
     *
     * @param  evt  A PropertyChangeEvent object describing the event source
     *              and the property that has changed.
     */
    public void propertyChange(PropertyChangeEvent evt) {
        if (evt.getPropertyName().equals("fileOpened")) {
            Object o = evt.getNewValue();
            if (o instanceof FileSource) {
                FileSource fs = (FileSource) o;
                // Is this file already in our list?
                FileItemPair fip = findPair(fs);
                if (fip == null) {

                    // Make a new menu item for this file.
                    String filepath = null;
                    try {
                        filepath = fs.getFile().getCanonicalPath();
                    } catch (IOException ioe) {
                        // This is highly unlikely.
                        return;
                    }
                    JMenuItem item = createMenuItem(filepath);
                    fip = new FileItemPair(fs, item);

                    // Trim the file-item list.
                    int maxSize = preferences.getInt("maxListSize",
                                                     Defaults.MRO_LIST_SIZE);
                    recentFiles.add(fip);
                    while (recentFiles.size() > maxSize) {
                        recentFiles.remove(0);
                    }

                    // Trim the menu down to size.
                    while (getMenuComponentCount() > maxSize) {
                        remove(getMenuComponentCount() - 1);
                    }

                    // Update the preferences information.
                    for (int ii = maxSize - 1, jj = maxSize;
                         ii >= 1; ii--, jj--) {
                        String val = preferences.get(NODE_PREFIX + ii, null);
                        if (val != null) {
                            preferences.put(NODE_PREFIX + jj, val);
                        }
                    }
                    // Now store the new file path.
                    preferences.put(NODE_PREFIX + '1', filepath);
                }
            }
        }
    } // propertyChange

    /**
     * A pairing of a FileSource and a JMenuItem.
     *
     * @author  Nathan Fiedler
     */
    protected class FileItemPair {
        /** The file source. */
        private FileSource file;
        /** The menu item. */
        private JMenuItem item;

        /**
         * Constructs a file and item pairing.
         *
         * @param  file  file source.
         * @param  item  menu item.
         */
        public FileItemPair(FileSource file, JMenuItem item) {
            this.file = file;
            this.item = item;
        } // FileItemPair

        /**
         * Returns the file.
         *
         * @return  the file.
         */
        public FileSource getFileSource() {
            return file;
        } // getFileSource

        /**
         * Returns the menu item.
         *
         * @return  the menu item.
         */
        public JMenuItem getMenuItem() {
            return item;
        } // getMenuItem
    } // FileItemPair
} // OpenedFilesMenu
