/*********************************************************************
 *
 *      Copyright (C) 2001-2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Utilities
 * FILE:        ClassUtils.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      05/19/01        Initial version
 *      nf      03/08/02        Added findMethod() method
 *      nf      03/18/02        Added callToString() method
 *      nf      03/20/02        Added isPrimitive() method
 *      nf      03/22/02        Added fuzzy-logic to findMethod()
 *      nf      04/17/02        Using new preferences facility
 *      nf      06/13/02        Added invokeMethod() method
 *      nf      08/10/02        Added isMethodIdentifier() method
 *      nf      12/28/02        Fixed bug 576
 *
 * $Id: ClassUtils.java 916 2003-04-06 18:44:57Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.util;

import com.bluemarsh.jswat.Defaults;
import com.bluemarsh.jswat.breakpoint.AmbiguousClassSpecException;
import com.bluemarsh.jswat.breakpoint.AmbiguousMethodException;
import com.sun.jdi.ClassType;
import com.sun.jdi.IncompatibleThreadStateException;
import com.sun.jdi.Method;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.Value;
import com.sun.jdi.VirtualMachine;
import java.io.InputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.prefs.Preferences;

/**
 * Class ClassUtils provides a set of utility functions for dealing with
 * classes and their members.
 *
 * @author  Nathan Fiedler
 */
public class ClassUtils {
    /** Our Preferences node. */
    private static Preferences preferences;

    static {
        preferences = Preferences.userRoot().node("com/bluemarsh/jswat/util");
    }

    /**
     * Call the <code>toString()</code> method on the given object.
     * <em>Note:</em> This method invalidates any
     * <code>StackFrame</code> instances for the given thread. The
     * caller will have to retrieve a new instance of
     * <code>StackFrame</code>, as needed.
     *
     * @param  obj     object on which to call toString().
     * @param  thread  thread on which to call toString() on obj.
     * @return  Return value from <code>obj.toString()</code>, or
     *          null if there was any problem.
     * @throws  Exception
     *          just about anything could go wrong.
     */
    public static String callToString(ObjectReference obj,
                                      ThreadReference thread)
        throws Exception {
        try {
            ReferenceType type = obj.referenceType();
            // Find the zero-arg toString() method that returns a String.
            List methods = type.methodsByName(
                "toString", "()Ljava/lang/String;");
            // First one better be the right one.
            Method method = (Method) methods.get(0);
            // Must invoke single-threaded or other threads will run
            // briefly and strange things will seem to happen from
            // the debugger user's point of view.
            int timeout = preferences.getInt("invocationTimeout",
                                             Defaults.INVOCATION_TIMEOUT);
            Object o = invokeMethod(obj, type, thread, method,
                                    new LinkedList(), timeout,
                                    ObjectReference.INVOKE_SINGLE_THREADED);
            String s = o == null ? "null" : o.toString();
            return StringUtils.trimQuotes(s);
        } catch (IncompatibleThreadStateException itse) {
            // This may happen if the thread is not suspended by JDI.
            return null;
        }
    } // callToString

    /**
     * Return a list of classes and interfaces whose names match the
     * given pattern. The pattern syntax is a fully-qualified class name
     * in which the first part or last part may optionally be a "*"
     * character, to match any sequence of characters.
     *
     * @param  vm       debuggee virtual machine.
     * @param  pattern  Classname pattern, optionally prefixed or
     *                  suffixed with "*" to match anything.
     * @return  List of ReferenceType objects.
     */
    public static List findClassesByPattern(VirtualMachine vm,
                                            String pattern) {

        List result = new ArrayList();
        if (pattern.indexOf('*') == -1) {
            // It's just a class name, try to find it.
            return vm.classesByName(pattern);
        } else {
            // Wild card exists, have to search manually.
            boolean head = true;
            if (pattern.startsWith("*")) {
                pattern = pattern.substring(1);
            } else if (pattern.endsWith("*")) {
                pattern = pattern.substring(0, pattern.length() - 1);
                head = false;
            }
            vm.suspend();
            List classes = vm.allClasses();
            vm.resume();
            Iterator iter = classes.iterator();
            while (iter.hasNext()) {
                ReferenceType type = (ReferenceType) iter.next();
                if (head && type.name().endsWith(pattern)) {
                    result.add(type);
                } else if (!head && type.name().startsWith(pattern)) {
                    result.add(type);
                }
            }
            return result;
        }
    } // findClassesByPattern

    /**
     * Attempt an unambiguous match of the method name and argument
     * specification to a method. If no arguments are specified, the
     * method must not be overloaded. Otherwise, the argument types must
     * match exactly.
     *
     * @param  clazz       class in which to find method.
     * @param  methodId    name of method to find.
     * @param  methodArgs  list of method argument types.
     * @return  Method if found, null otherwise.
     * @exception  AmbiguousClassSpecException
     *             Thrown if the given pattern matches more than one class.
     * @exception  AmbiguousMethodException
     *             Thrown if the method is overloaded.
     * @exception  NoSuchMethodException
     *             Thrown if the method could not be found.
     */
    public static Method findMethod(ReferenceType clazz,
                                    String methodId,
                                    List methodArgs)
        throws AmbiguousClassSpecException,
               AmbiguousMethodException,
               NoSuchMethodException {

        // This ought to search only the methods of the given class. The
        // caller should then invoke this method again with the
        // superclass until a match is found.

        // Normalize the argument list before comparing.
        List argTypeNames = null;
        if (methodArgs != null) {
            argTypeNames = new ArrayList(methodArgs.size());
            int size = methodArgs.size();
            for (int i = 0; i < size; i++) {
                String name = (String) methodArgs.get(i);
                if (!isPrimitive(name)) {
                    name = normalizeArgTypeName(name,
                                                clazz.virtualMachine());
                }
                argTypeNames.add(name);
            }
        }

        List methods = clazz.methodsByName(methodId);
        int size = methods.size();
        // A method with matching name (and possibly signature).
        Method match = null;
        // The method has the exact signature.
        boolean exact = false;
        // More than one means the method is overloaded.
        int matchCount = 0;
        Iterator iter = methods.iterator();
        while (iter.hasNext()) {
            Method candidate = (Method) iter.next();
            if (candidate.isAbstract() || candidate.isConstructor()) {
                // We cannot invoke abstract methods or constructors.
                continue;
            }
            matchCount++;
            match = candidate;
            // Check if argument types are the same, if any are given.
            if (argTypeNames != null) {
                List candidateNames = match.argumentTypeNames();

                // Check that the number of arguments is the same.
                if (candidateNames.size() == argTypeNames.size()) {
                    // Assume the arguments will match.
                    boolean argsMatch = true;
                    // Compare the list of arguments by type name.
                    for (int ii = 0; ii < candidateNames.size(); ii++) {
                        String arg1 = (String) candidateNames.get(ii);
                        String arg2 = (String) argTypeNames.get(ii);
                        if (!arg1.equals(arg2)) {
                            // See if we can match the arguments using
                            // fuzzy logic. First try the primitives
                            // (e.g. "int" == "java.lang.Integer"),
                            // then try up-casting the types.
                            if (!tryPrimitives(arg1, arg2)
                                && !tryUpCasting(arg1, arg2)) {
                                argsMatch = false;
                                break;
                            }
                        }
                    }
                    // Found a matching method.
                    if (argsMatch) {
                        exact = true;
                        break;
                    }
                }
            }
        }

        if (exact) {
            return match;
        } else if ((argTypeNames == null) && (matchCount > 0)) {
            if (matchCount == 1) {
                // There's only one match, good enough for us.
                return match;
            } else {
                // No args were given to make an exact match.
                throw new AmbiguousMethodException();
            }
        } else {
            throw new NoSuchMethodException(methodId);
        }
    } // findMethod

    /**
     * Find the ReferenceType object for a given class.
     *
     * @param  cname  Class identifier for which to get reference.
     * @param  vm     Virtual machine from which to get reference.
     * @return  ReferenceType for this class, or null if none.
     * @exception  AmbiguousClassSpecException
     *             Thrown if the given pattern matches more than one class.
     */
    protected static ReferenceType getReferenceFromName(String cname,
                                                        VirtualMachine vm)
        throws AmbiguousClassSpecException {
        ReferenceType cls = null;
        if (cname.startsWith("*.")) {
            // It's a pattern match if ID starts with '*'.
            // The first loaded class whose name matches is selected.
            cname = cname.substring(1);
            vm.suspend();
            List classes = vm.allClasses();
            vm.resume();

            // Go through all of the classes looking for one that
            // matches what we were given.
            Iterator iter = classes.iterator();
            while (iter.hasNext()) {
                ReferenceType type = ((ReferenceType) iter.next());
                // Check only the end of the ID, for pattern match.
                if (type.name().endsWith(cname)) {
                    cls = type;
                    break;
                }
            }
        } else {
            // It's the whole class name, that's easy.
            List classes = vm.classesByName(cname);
            // We don't care if this is an interface or array.
            if (classes.size() > 0) {
                if (classes.size() > 1) {
                    // Oops, more than one class matched.
                    throw new AmbiguousClassSpecException();
                }
                cls = (ReferenceType) classes.get(0);
            }
        }
        return cls;
    } // getReferenceFromName

    /**
     * Perform the hotswap operation on the given class. If anything
     * goes wrong, it will be reported as an exception.
     *
     * @param  clazz  the class to be hotswapped.
     * @param  code   the bytecode to replace the class.
     * @param  vm     virtual machine.
     * @throws  ClassCircularityError
     *          if the class dependencies are circular.
     * @throws  ClassFormatError
     *          if the class format is wrong.
     * @throws  IOException
     *          if there was an error reading the input stream.
     * @throws  NoClassDefFoundError
     *          if the class definition was not found.
     * @throws  UnsupportedClassVersionError
     *          if the class is the wrong version.
     * @throws  UnsupportedOperationException
     *          if the hotswap operation is not supported.
     * @throws  VerifyError
     *          if the bytecode verfication failed.
     */
    public static void hotswap(ReferenceType clazz, InputStream code,
                               VirtualMachine vm)
        throws ClassCircularityError,
               ClassFormatError,
               IOException,
               NoClassDefFoundError,
               UnsupportedClassVersionError,
               UnsupportedOperationException,
               VerifyError {

        // Load the byte-code from the class file.
        byte[] byteCode = new byte[1024];
        int totalBytesRead = 0;
        int length = byteCode.length;
        while (true) {
            int bytesRead = code.read(byteCode, totalBytesRead, length);
            if (bytesRead == -1) {
                break;
            }

            totalBytesRead += bytesRead;
            if (totalBytesRead >= byteCode.length) {
                byte[] temp = new byte[totalBytesRead * 2];
                System.arraycopy(byteCode, 0, temp, 0, byteCode.length);
                byteCode = temp;
            }
            length = byteCode.length - totalBytesRead;
        }
        byte[] temp = new byte[totalBytesRead];
        System.arraycopy(byteCode, 0, temp, 0, temp.length);
        byteCode = temp;

        // Do the actual hotswap operation.
        Map map = new HashMap();
        map.put(clazz, byteCode);
        vm.redefineClasses(map);
    } // hotswap

    /**
     * Invokes the given method on the class or object. If the method
     * call times out, this method returns <code>null</code>.
     *
     * @param  objref      object reference, if non-static method.
     * @param  reftype     reference type, if static method.
     * @param  thread      thread on which to invoke method.
     * @param  method      method to be invoked.
     * @param  methodArgs  arguments to the method.
     * @param  timeout     milliseconds to wait for method to return.
     * @return  the return value of the method.
     * @throws  Exception
     *          if something goes wrong.
     */
    public static Object invokeMethod(ObjectReference objref,
                                      ReferenceType reftype,
                                      ThreadReference thread,
                                      Method method,
                                      List methodArgs,
                                      int timeout) throws Exception {
        return invokeMethod(objref, reftype, thread, method, methodArgs,
                            timeout, 0);
    } // invokeMethod

    /**
     * Invokes the given method on the class or object. If the method
     * call times out, this method returns <code>null</code>.
     *
     * @param  objref      object reference, if non-static method.
     * @param  reftype     reference type, if static method.
     * @param  thread      thread on which to invoke method.
     * @param  method      method to be invoked.
     * @param  methodArgs  arguments to the method.
     * @param  timeout     milliseconds to wait for method to return.
     * @param  flags       flags to pass to JPDA's invokeMethod().
     * @return  the return value of the method.
     * @throws  Exception
     *          if something goes wrong.
     */
    public static Object invokeMethod(ObjectReference objref,
                                      ReferenceType reftype,
                                      ThreadReference thread,
                                      Method method, List methodArgs,
                                      int timeout, int flags)
        throws Exception {

        Invoker invoker = new Invoker(objref, reftype, thread, method,
                                      methodArgs, flags);
        Thread th = new Thread(invoker);
        th.start();
        synchronized (invoker) {
            try {
                invoker.wait(timeout);
            } catch (InterruptedException ie) {
                // ignored
            }
        }
        Exception excp = invoker.getException();
        if (excp != null) {
            throw excp;
        } else {
            return invoker.getValue();
        }
    } // invokeMethod

    /**
     * Determine if the given string is a valid Java identifier.
     *
     * @param  s  string to validate.
     * @return  true if string is a valid Java identifier.
     */
    public static boolean isJavaIdentifier(String s) {
        if (s == null || s.length() == 0) {
            return false;
        }
        // First character of identifier is a special case.
        if (!Character.isJavaIdentifierStart(s.charAt(0))) {
            return false;
        }
        // Now check all other characters of the identifier.
        for (int i = 1; i < s.length(); i++) {
            if (!Character.isJavaIdentifierPart(s.charAt(i))) {
                return false;
            }
        }
        return true;
    } // isJavaIdentifier

    /**
     * Determine if the given string is a valid method identifier.
     *
     * @param  s  string to validate.
     * @return  true if string is a valid method identifier.
     */
    public static boolean isMethodIdentifier(String s) {
        return isJavaIdentifier(s) || s.equals("<init>")
            || s.equals("<clinit>");
    } // isMethodIdentifier

    /**
     * Returns true if the passed string is a primitive type.
     *
     * @param  s  string suspected of naming a primitive type.
     * @return  true if 's' names a primitive, false otherwise.
     */
    public static boolean isPrimitive(String s) {
        if (s == null || s.length() == 0) {
            return false;
        }

        if (s.equals("boolean") || s.equals("byte") || s.equals("char")
            || s.equals("double") || s.equals("float") || s.equals("int")
            || s.equals("long") || s.equals("short")) {
            return true;
        } else {
            return false;
        }
    } // isPrimitive

    /**
     * Returns just the name of the class, without the package name.
     *
     * @param  cname  Name of class, possibly fully-qualified.
     * @return  Just the class name, or null if cname is null.
     */
    public static String justTheName(String cname) {
        if (cname == null) {
            return null;
        }

        if (preferences.getBoolean("shortClassNames",
                                   Defaults.SHORT_CLASS_NAMES)) {
            int i = cname.lastIndexOf('.');
            if (i > 0) {
                return cname.substring(i + 1);
            }
        }
        return cname;
    } // justTheName

    /**
     * Remove unneeded spaces and expand class names to fully qualified
     * names, if necessary and possible.
     *
     * @param  name  Name of class.
     * @param  vm    Virtual machine to normalize against.
     * @return  Fully qualified name of class.
     * @exception  AmbiguousClassSpecException
     *             Thrown if the given pattern matches more than one class.
     */
    protected static String normalizeArgTypeName(String name,
                                                 VirtualMachine vm)
        throws AmbiguousClassSpecException {
        name = name.trim();
        // Separate type name and any array modifiers.
        StringBuffer className = new StringBuffer();
        StringBuffer arrayPart = new StringBuffer();

        // Get the name of the class (the part before whitespace or [).
        int i = 0;
        while (i < name.length()) {
            char c = name.charAt(i);
            if (Character.isWhitespace(c) || (c == '[')) {
                break;
            }
            className.append(c);
            i++;
        }

        // Get the array portion of the name (just [] with whitespace).
        while (i < name.length()) {
            char c = name.charAt(i);
            if ((c == '[') || (c == ']')) {
                arrayPart.append(c);
            } else if (!Character.isWhitespace(c)) {
                throw new IllegalArgumentException("Invalid array type");
            }
            i++;
        }

        // If there's no package name given, try to make the class
        // name fully qualified.
        name = className.toString();
        if ((name.indexOf('.') == -1) || name.startsWith("*.")) {
            ReferenceType argClass = getReferenceFromName(name, vm);
            if (argClass != null) {
                name = argClass.name();
            }
        }
        return name + arrayPart.toString();
    } // normalizeArgTypeName

    /**
     * Compare the named types to determine if they are compatible by
     * one being a class type representing the primitive form of the
     * other. That is, if one argument is a primitive type and the other
     * is the class type that represents that primitive, return true.
     * Otherwise, return false.
     *
     * @param  candidate  type we are trying to match.
     * @param  given      type the user specified, which may be up-cast.
     * @return  true if types are equivalent, false otherwise.
     */
    protected static boolean tryPrimitives(String candidate, String given) {
        if (candidate.equals("byte") && given.equals("java.lang.Byte")
            || candidate.equals("short") && given.equals("java.lang.Short")
            || candidate.equals("int") && given.equals("java.lang.Integer")
            || candidate.equals("long") && given.equals("java.lang.Long")
            || candidate.equals("float") && given.equals("java.lang.Float")
            || candidate.equals("double") && given.equals("java.lang.Double")
            || candidate.equals("boolean") && given.equals("java.lang.Boolean")
            || candidate.equals("char")
            && given.equals("java.lang.Character")) {
            return true;
        } else {
            return false;
        }
    } // tryPrimitives

    /**
     * Compare the named types to determine if up-casting is possible.
     * Note that this only works with class types that represent
     * primitive numeric types (byte, short, integer, long, float,
     * double).
     *
     * @param  candidate  type we are trying to match.
     * @param  given      type the user specified, which may be up-cast.
     * @return  true if up-cast is possible, false otherwise.
     */
    protected static boolean tryUpCasting(String candidate, String given) {
        // Granted, string comparisons aren't exactly quick, but then
        // we hardly expect this to be invoked very often.
        if (candidate.equals("java.lang.Short")
            || candidate.equals("short")) {
            if (given.equals("java.lang.Byte")
                || given.equals("byte")) {
                return true;
            }
        } else if (candidate.equals("java.lang.Integer")
                   || candidate.equals("int")) {
            if (given.equals("java.lang.Byte")
                || given.equals("java.lang.Short")
                || given.equals("byte")
                || given.equals("short")) {
                return true;
            }
        } else if (candidate.equals("java.lang.Long")
                   || candidate.equals("long")) {
            if (given.equals("java.lang.Byte")
                || given.equals("java.lang.Short")
                || given.equals("java.lang.Integer")
                || given.equals("byte")
                || given.equals("short")
                || given.equals("integer")) {
                return true;
            }
        } else if (candidate.equals("java.lang.Double")
                   || candidate.equals("double")) {
            if (given.equals("java.lang.Float")
                || given.equals("float")) {
                return true;
            }
        }
        return false;
    } // tryUpCasting

    /**
     * Invokes a method and returns the result.
     */
    protected static class Invoker implements Runnable {
        /** Return value from invoked method. */
        private Value value;
        /** The exception thrown, if any. */
        private Exception exception;
        /** Object reference, if non-static method. */
        private ObjectReference objref;
        /** Reference type, if static method. */
        private ReferenceType reftype;
        /** Thread on which to invoke method. */
        private ThreadReference thread;
        /** Method to be invoked. */
        private Method method;
        /** Arguments to the method. */
        private List methodArgs;
        /** Flags to pass to invokeMethod(). */
        private int flags;

        /**
         * Constructs an Invoker to invoke the given method.
         *
         * @param  objref      object reference, if non-static method.
         * @param  reftype     reference type, if static method.
         * @param  thread      thread on which to invoke method.
         * @param  method      method to be invoked.
         * @param  methodArgs  arguments to the method.
         * @param  flags       flags to pass to invokeMethod().
         */
        public Invoker(ObjectReference objref, ReferenceType reftype,
                       ThreadReference thread, Method method,
                       List methodArgs, int flags) {
            this.objref = objref;
            this.reftype = reftype;
            this.thread = thread;
            this.method = method;
            this.methodArgs = methodArgs;
            this.flags = flags;
        } // Invoker

        /**
         * Returns the exception that occurred while invoking the
         * method, or null if no exception occurred.
         *
         * @return  thrown exception.
         */
        public Exception getException() {
            return exception;
        } // getException

        /**
         * Returns the return value of the invoked method.
         *
         * @return  method return value.
         */
        public Value getValue() {
            return value;
        } // getValue

        /**
         * Invoke the desired method.
         */
        public void run() {
            // Clear these in case we are called more than once.
            value = null;
            exception = null;
            try {
                if (objref == null) {
                    // Static method invocation.
                    if (reftype instanceof ClassType) {
                        ClassType clazz = (ClassType) reftype;
                        value = clazz.invokeMethod(thread, method,
                                                   methodArgs, flags);
                    } else {
                        exception = new IllegalArgumentException(
                            "Specified class is an array or interface.");
                    }

                } else {
                    // Non-static method invocation.
                    value = objref.invokeMethod(thread, method, methodArgs,
                                                flags);
                }
                synchronized (this) {
                    notify();
                }
            } catch (Exception e) {
                exception = e;
            }
        } // run
    } // Invoker
} // ClassUtils
