/*********************************************************************
 *
 *      Copyright (C) 2001-2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Utilities
 * FILE:        FieldAndValue.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      09/03/01        Initial version
 *      nf      07/21/02        Added array.length support
 *
 * $Id: FieldAndValue.java 962 2003-05-22 04:53:57Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.util;

import com.sun.jdi.ArrayReference;
import com.sun.jdi.ClassNotLoadedException;
import com.sun.jdi.Field;
import com.sun.jdi.LocalVariable;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.Value;
import com.sun.jdi.Type;

/**
 * Encapsulates a Field, its Value, and the Object containing the Field.
 * May also represent a local variable, so the field will be null, and
 * the object will be the LocalVariable. Additionally, this wrapper may
 * represent an array .length reference.
 *
 * @author  Nathan Fiedler
 */
public class FieldAndValue {
    /** Field reference, or null if local variable. */
    private Field field;
    /** Value of Field (always non-null). */
    private Value value;
    /** Object containing field, or null if static field or local variable. */
    private ObjectReference object;
    /** Local variable, if object and field are null. */
    private LocalVariable localVar;

    /**
     * Constructs a FieldAndValue to hold a field and its value.
     *
     * @param  field   Field reference.
     * @param  value   Value of Field.
     * @param  object  ObjectReference.
     */
    public FieldAndValue(Field field, Value value, ObjectReference object) {
        this.field = field;
        this.value = value;
        this.object = object;
    } // FieldAndValue

    /**
     * Constructs a FieldAndValue to hold an array reference.
     *
     * @param  field   Field reference.
     * @param  array   ArrayReference.
     */
    public FieldAndValue(Field field, ArrayReference array) {
        // Value is the array length.
        this(field, array.virtualMachine().mirrorOf(array.length()), array);
    } // FieldAndValue

    /**
     * Constructs a FieldAndValue to hold a local variable and
     * its value.
     *
     * @param  local  LocalVariable.
     * @param  value  Value of LocalVariable.
     */
    public FieldAndValue(LocalVariable local, Value value) {
        this.localVar = local;
        this.value = value;
    } // FieldAndValue

    /**
     * Constructs a FieldAndValue to hold an array reference.
     *
     * @param  local  LocalVariable.
     * @param  array   ArrayReference.
     */
    public FieldAndValue(LocalVariable local, ArrayReference array) {
        // Value is the array length.
        this(local, array.virtualMachine().mirrorOf(array.length()));
    } // FieldAndValue

    /**
     * Returns the Field object, if this represents a field.
     *
     * @return  field.
     */
    public Field getField() {
        return field;
    } // getField

    /**
     * Returns the LocalVariable object, if this represents a local
     * variable.
     *
     * @return  local variable, or null if field.
     */
    public LocalVariable getLocalVar() {
        return localVar;
    } // getLocalVar

    /**
     * Returns the ObjectReference that contains the field, if this
     * represents a non-static field.
     *
     * @return  object, or null if local variable or static field.
     */
    public ObjectReference getObject() {
        return object;
    } // getObject

    /**
     * Type of the value if non-null, or the type of the field or local
     * variable if the value is null.
     *
     * @return  type of value.
     */
    public Type getType() {
        try {
            if (value != null) {
                return value.type();
            } else if (field != null) {
                return field.type();
            } else if (localVar != null) {
                return localVar.type();
            } else {
                return null;
            }
        } catch (ClassNotLoadedException cnle) {
            return null;
        }
    } // getType

    /**
     * Returns the Value object.
     *
     * @return  value, or null if local variable.
     */
    public Value getValue() {
        return value;
    } // getValue
} // FieldAndValue
