/*********************************************************************
 *
 *      Copyright (C) 2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      View
 * FILE:        JavaSourceView.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      01/30/03        Initial version
 *      nf      03/06/03        Fixed bug 728
 *
 * $Id: JavaSourceView.java 953 2003-04-29 06:06:52Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.view;

import com.bluemarsh.jswat.ContextManager;
import com.bluemarsh.jswat.Defaults;
import com.bluemarsh.jswat.event.ContextChangeEvent;
import com.bluemarsh.jswat.event.ContextListener;
import com.bluemarsh.jswat.event.SessionEvent;
import com.bluemarsh.jswat.parser.java.lexer.LexerException;
import com.bluemarsh.jswat.ui.UIAdapter;
import com.bluemarsh.jswat.util.SkipList;
import com.sun.jdi.Location;
import java.io.CharArrayReader;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;

/**
 * Defines the JavaSourceView class which extends SourceView and displays
 * Java source code.
 *
 * @author  Nathan Fiedler
 */
class JavaSourceView extends SourceView implements ContextListener {
    /** Draw layer responsible for colorizing Java code. */
    private JavaDrawLayer javaDrawLayer;
    /** Classlines of the parsed source file. */
    private List classDefinitions;
    /** List of method definitions. */
    private List methodDefinitions;

    /**
     * Creates a JavaSourceView object.
     *
     * @param  title  title of the view.
     * @param  file   source file, if known.
     */
    public JavaSourceView(String title, File file) {
        super(title, file);

        // Create the text component.
        javaDrawLayer = new JavaDrawLayer();
        textComponent.addDrawLayer(javaDrawLayer,
                                   DrawLayer.PRIORITY_COLORIZER);

        // Remove the generic popup menu.
        textComponent.removeMouseListener(popupMenu);

        // Create and add our Java popup menu.
        popupMenu = new JavaSourceViewPopup(file, this);
        textComponent.addMouseListener(popupMenu);
    } // JavaSourceView

    /**
     * Called when the Session has activated. This occurs when the
     * debuggee has launched or has been attached to the debugger.
     *
     * @param  sevt  session event.
     */
    public void activated(SessionEvent sevt) {
        // Register for debugger events.
        ContextManager cmgr = (ContextManager)
            sevt.getSession().getManager(ContextManager.class);
        cmgr.addContextListener(this);
    } // activated

    /**
     * Invoked when the current context has changed. The context change
     * event identifies which aspect of the context has changed.
     *
     * @param  cce  context change event
     */
    public void contextChanged(ContextChangeEvent cce) {
        ContextManager cmgr = (ContextManager) cce.getSource();
        Location loc = cmgr.getCurrentLocation();
        // See if this event belongs to our source file.
        if (loc != null && matches(loc)) {
            int line = loc.lineNumber();
            // Show where the pc is located.
            // Minus one for some reason.
            showHighlight(line - 1);
            scrollToLine(line);
        } else {
            // Clear the current stepping line.
            removeHighlight();
        }
    } // contextChanged

    /**
     * Called when the Session has deactivated. The debuggee VM is no
     * longer connected to the Session.
     *
     * @param  sevt  session event.
     */
    public void deactivated(SessionEvent sevt) {
        ContextManager cmgr = (ContextManager)
            sevt.getSession().getManager(ContextManager.class);
        cmgr.removeContextListener(this);
        // Clear the current stepping line.
        removeHighlight();
        if (logger.isLoggable(Level.INFO)) {
            logger.info("setting current line to 0 in " + viewTitle);
        }
    } // deactivated

    /**
     * Check if the Location is in the same source file as the file we
     * are displaying.
     *
     * @param  location  Location to check.
     * @return  True if breakpoint source name same as our filename.
     */
    protected boolean matches(Location location) {
        if (classDefinitions == null) {
            return false;
        } else {
            Iterator iter = classDefinitions.iterator();
            String clazz = location.declaringType().name();
            while (iter.hasNext()) {
                ClassDefinition cd = (ClassDefinition) iter.next();
                if (cd.getClassName().equals(clazz)) {
                    return true;
                }
            }
        }
        return false;
    } // matches

    /**
     * Reads the class definition information from the source file using
     * the JavaParser class. Give this information to the text component
     * and row header popup menus.
     */
    protected void parseClassDefs() {
        // Read in the class definitions from the source file.
        CharArrayReader car = new CharArrayReader(viewContent);
        JavaParser parser = new JavaParser(car);
        try {
            parser.parse();
            // Parser closes the reader in all cases.
            classDefinitions = parser.getClassLines();
            methodDefinitions = parser.getMethodLines();
        } catch (Exception e) {
            // Don't spew an entire stack trace.
            owningSession.getUIAdapter().showMessage(UIAdapter.MESSAGE_ERROR,
                                                     e.toString());
            classDefinitions = null;
            methodDefinitions = null;
        }
        if (logger.isLoggable(Level.INFO)) {
            logger.info("parsed Java source in " + viewTitle);
        }
        breakpointDrawLayer.setClassDefs(classDefinitions);
        JavaSourceViewPopup popup = (JavaSourceViewPopup) popupMenu;
        popup.setClassDefinitions(classDefinitions, methodDefinitions);
    } // parseClassDefs

    /**
     * Code to be executed after the refresh() method has completed.
     */
    protected void refreshPost() {
        // Parse first so the breakpoints can use the information.
        boolean parse = preferences.getBoolean("parseView",
                                               Defaults.VIEW_PARSE);
        if (parse) {
            parseClassDefs();
        }

        // Set all the breakpoints and the current line marker.
        ContextManager cmgr = (ContextManager)
            owningSession.getManager(ContextManager.class);
        Location loc = cmgr.getCurrentLocation();
        if (loc != null && matches(loc)) {
            // Show where the pc is located.
            // Minus one for some reason.
            showHighlight(loc.lineNumber() - 1);
        }

        // Now call the superclass so the draw layers get the class
        // definitions they need.
        super.refreshPost();
    } // refreshPost

    /**
     * Called to update this view's preferences, either when this object
     * is constructed or when the preferences change.
     */
    protected void setPreferences() {
        super.setPreferences();
        // Notify the colorizers of possible changes.
        try {
            JavaDrawLayer.updateColors(preferences);
        } catch (NumberFormatException nfe) {
            owningSession.getUIAdapter().showMessage(UIAdapter.MESSAGE_ERROR,
                                                     nfe.toString());
        }
        // Short method descriptor option may have changed.
        JavaSourceViewPopup popup = (JavaSourceViewPopup) popupMenu;
        if (popup != null) {
            popup.setClassDefinitions(classDefinitions, methodDefinitions);
        }
    } // setPreferences

    /**
     * Set the content of the text component using the text defined by
     * <code>viewContent</code>. This implementation applies styles to
     * the Java source to colorize it syntactically.
     */
    protected void setTextContent() {
        boolean colorize = preferences.getBoolean("colorizeView",
                                                  Defaults.VIEW_COLORIZE);

        SourceContent doc = new SourceContent(viewContent);

        if (colorize) {
            // Colorize the source.
            CharArrayReader car = new CharArrayReader(viewContent);
            JavaScanner scanner = new JavaScanner(car, doc);
            SkipList tokenInfo = null;
            try {
                tokenInfo = scanner.scan();
            } catch (IOException ioe) {
                owningSession.getUIAdapter().showMessage(
                    UIAdapter.MESSAGE_ERROR, ioe.toString());
            } catch (LexerException le) {
                owningSession.getUIAdapter().showMessage(
                    UIAdapter.MESSAGE_ERROR, le.toString());
            }
            // Scanner closes the reader.
            // Give the JavaDrawLayer the token information.
            javaDrawLayer.setTokens(tokenInfo);
        } else {
            // Indicate that we have no tokens and the layer
            // should be inactive.
            javaDrawLayer.setTokens(null);
        }

        // Do this when we're all done, so it forces a repaint.
        textComponent.setContent(doc);
    } // setTextContent
} // JavaSourceView
