/*********************************************************************
 *
 *      Copyright (C) 2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      View
 * FILE:        SourceViewPopup.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      01/30/03        Initial version
 *      nf      01/30/03        Fixed bug 714
 *      nf      04/28/03        Fixed bug 764
 *
 * $Id: JavaSourceViewPopup.java 953 2003-04-29 06:06:52Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.view;

import com.bluemarsh.jswat.Defaults;
import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.breakpoint.Breakpoint;
import com.bluemarsh.jswat.breakpoint.BreakpointManager;
import com.bluemarsh.jswat.breakpoint.LineBreakpoint;
import com.bluemarsh.jswat.breakpoint.ResolveException;
import com.bluemarsh.jswat.breakpoint.ui.EditorDialog;
import com.bluemarsh.jswat.ui.SessionFrameMapper;
import com.bluemarsh.jswat.ui.UIAdapter;
import java.awt.Component;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.prefs.Preferences;
import javax.swing.JComponent;
import javax.swing.JMenu;
import javax.swing.JMenuItem;

/**
 * Class JavaSourceViewPopup extends SourceViewPopup and provides the
 * Java-specific features.
 *
 * @author  Nathan Fiedler
 */
class JavaSourceViewPopup extends SourceViewPopup {
    /** Name for add breakpoint menu item. */
    private static final String ADD_STRING = "addBreakpoint";
    /** Name for runto breakpoint menu item. */
    private static final String RUNTO_STRING = "runtoBreakpoint";
    /** Name for properties breakpoint menu item. */
    private static final String PROPS_STRING = "propsBreakpoint";
    /** Name for remove breakpoint menu item. */
    private static final String REMOVE_STRING = "removeBreakpoint";
    /** Name for enable breakpoint menu item. */
    private static final String ENABLE_STRING = "enableBreakpoint";
    /** Name for disable breakpoint menu item. */
    private static final String DISABLE_STRING = "disableBreakpoint";
    /** Name for "Java parser error" menu item. */
    private static final String PARSER_ERROR_STRING = "parserError";
    /** Name for "bad line number" menu item. */
    private static final String BAD_LINE_STRING = "badLine";
    /** Suffix added to command string to retrieve menu labels. */
    private static final String LABEL_SUFFIX = "Label";
    /** List of class definition objects. */
    private List classLines;
    /** If non-null, name of class user last right-clicked within. */
    private String lastClickedClass;
    /** If non-null, breakpoint user last right-clicked on. */
    private Breakpoint lastClickedBreakpoint;
    /** The "Methods" submenu. */
    private JMenu methodsMenu;
    /** Our parent view. */
    private SourceView sourceView;

    /**
     * Creates a JavaSourceViewPopup object.
     *
     * @param  file  file for source, if available.
     * @param  view  source view parent.
     */
    public JavaSourceViewPopup(File file, SourceView view) {
        super(file);
        sourceView = view;
        menuItemTable.put(ADD_STRING, createMenuItem(ADD_STRING));
        menuItemTable.put(RUNTO_STRING, createMenuItem(RUNTO_STRING));
        menuItemTable.put(PROPS_STRING, createMenuItem(PROPS_STRING));
        menuItemTable.put(REMOVE_STRING, createMenuItem(REMOVE_STRING));
        menuItemTable.put(ENABLE_STRING, createMenuItem(ENABLE_STRING));
        menuItemTable.put(DISABLE_STRING, createMenuItem(DISABLE_STRING));
        menuItemTable.put(PARSER_ERROR_STRING,
                          new JMenuItem(Bundle.getString(
                              "SourceViewPopup.parserErrorMsg")));
        menuItemTable.put(BAD_LINE_STRING,
                          new JMenuItem(Bundle.getString(
                              "SourceViewPopup.badLineNumMsg")));
    } // JavaSourceViewPopup

    /**
     * One of the menu items we're listening to was activated.
     *
     * @param  ae  action event.
     */
    public void actionPerformed(ActionEvent ae) {
        // Call the superclass to perform its options. Then we do
        // our own checks, but we won't duplicate the effort.
        super.actionPerformed(ae);
        // Get the source of the event (it is a JMenuItem).
        JMenuItem menuItem = (JMenuItem) ae.getSource();
        Session session = SessionFrameMapper.getSessionForEvent(ae);
        BreakpointManager bpman = (BreakpointManager)
            session.getManager(BreakpointManager.class);
        // Get the action command.
        String cmd = menuItem.getActionCommand();
        // Decide what to do.
        if (cmd.equals(ADD_STRING) || cmd.equals(RUNTO_STRING)) {

            // Handle the "Add breakpoint" menu item.
            Breakpoint bp = addBreakpoint(session, bpman);
            if (bp != null && cmd.equals(RUNTO_STRING)) {
                // Make this breakpoint a little special.
                bp.deleteOnExpire();
                bp.setExpireCount(1);
                // Resume the debuggee so we do the "run" part.
                if (session.isActive()) {
                    session.resumeVM(this, false, false);
                }
            }

        } else if (cmd.equals(PROPS_STRING)) {

            // Handle the "Properties..." menu item.
            Component invoker = getInvoker();
            Frame frame = SessionFrameMapper.getOwningFrame(invoker);
            EditorDialog editor = new EditorDialog(
                frame, lastClickedBreakpoint);
            editor.setLocationRelativeTo(frame);
            editor.show();
        } else if (cmd.equals(REMOVE_STRING)) {
            // Delete the breakpoint using the breakpoint manager.
            bpman.removeBreakpoint(lastClickedBreakpoint);
        } else if (cmd.equals(ENABLE_STRING)) {
            // Enable the breakpoint using the breakpoint manager.
            bpman.enableBreakpoint(lastClickedBreakpoint);
        } else if (cmd.equals(DISABLE_STRING)) {
            // Disable the breakpoint using the breakpoint manager.
            bpman.disableBreakpoint(lastClickedBreakpoint);

        } else {

            // It must be one of the method menu items.
            try {
                int line = Integer.parseInt(cmd);
                sourceView.scrollToLine(line);
            } catch (NumberFormatException nfe) {
                // Or maybe not...
            }
        }
    } // actionPerformed

    /**
     * Try to add a breakpoint at the last clicked class and line.
     * Displays an appropriate error message if needed.
     *
     * @param  session  Session.
     * @param  bpman    breakpoint manager.
     * @return  new breakpoint if successful, null if error.
     */
    protected Breakpoint addBreakpoint(Session session,
                                       BreakpointManager bpman) {
        try {
            Breakpoint bp = new LineBreakpoint(lastClickedClass,
                                               lastClickedLine);
            bpman.addNewBreakpoint(bp);
            return bp;
        } catch (ClassNotFoundException cnfe) {
            session.getUIAdapter().showMessage(
                UIAdapter.MESSAGE_ERROR,
                Bundle.getString("AddBreak.invalidClassMsg") + ' '
                + lastClickedClass);
            return null;
        } catch (ResolveException re) {
            session.getUIAdapter().showMessage(
                UIAdapter.MESSAGE_ERROR, re.errorMessage());
            return null;
        }
    } // addBreakpoint

    /**
     * Finds the class defined at this line number.
     *
     * @param  line  Line number.
     * @return  Class name, or null if line not in a class.
     */
    protected String findClassForLine(int line) {
        int begin = Integer.MIN_VALUE;
        int end = Integer.MAX_VALUE;
        String cname = null;
        Iterator iter = classLines.iterator();
        // Go through the entire list to find the closest match.
        while (iter.hasNext()) {
            ClassDefinition cd = (ClassDefinition) iter.next();
            if ((cd.getBeginLine() <= line) && (cd.getEndLine() >= line)
                && (cd.getBeginLine() > begin) && (cd.getEndLine() < end)) {
                cname = cd.getClassName();
                begin = cd.getBeginLine();
                end = cd.getEndLine();
            }
        }
        return cname;
    } // findClassForLine

    /**
     * Invoked when a mouse button has been released on a component. We
     * use this opportunity to show the popup menu.
     *
     * @param  e  Mouse event.
     */
    public void mouseReleased(MouseEvent e) {
        if (!e.isConsumed()) {
            // Is this within the gutter area and the popup is not visible?
            SourceViewTextArea textArea = (SourceViewTextArea) e.getSource();
            if (textArea.isWithinGutter(e.getX(), e.getY())
                && !isVisible()) {
                // Yes, create/disable/delete breakpoint at this line.

                if (classLines == null) {
                    // Parser error when reading source file.
                    return;
                }

                // Get the class at this line.
                lastClickedLine = textArea.viewToLine(e.getPoint());
                lastClickedClass = findClassForLine(lastClickedLine);
                if (lastClickedClass == null) {
                    return;
                }

                // Query existing set of breakpoints for this class and line.
                Session session = SessionFrameMapper.getSessionForEvent(e);
                BreakpointManager bpman = (BreakpointManager)
                    session.getManager(BreakpointManager.class);
                lastClickedBreakpoint = bpman.getBreakpoint(
                    lastClickedClass, lastClickedLine);
                if (lastClickedBreakpoint == null) {
                    // No breakpoint at this location, add a new one.
                    addBreakpoint(session, bpman);
                } else {
                    // Breakpoint exists.
                    if (lastClickedBreakpoint.isEnabled()) {
                        // Breakpoint is enabled, disable it.
                        bpman.disableBreakpoint(lastClickedBreakpoint);
                    } else {
                        // Breakpoint is disabled, remove it.
                        bpman.removeBreakpoint(lastClickedBreakpoint);
                    }
                }
            } else {
                super.mouseReleased(e);
            }
        }
    } // mouseReleased

    /**
     * Remove all of the children, then put the methods submenu back in,
     * if it is defined.
     */
    protected void removeAllItems() {
        removeAll();
        // Add the ever-present menu items.
        add((JMenuItem) menuItemTable.get(EDITOR_STRING));
        if (methodsMenu != null) {
            add(methodsMenu);
        }
    } // removeAllItems

    /**
     * Set the lists of class and method definitions.
     *
     * @param  classes  list of ClassDefinition objects.
     * @param  methods  list of MethodDefinition objects.
     */
    void setClassDefinitions(List classes, List methods) {
        classLines = classes;
        if (classLines == null || methods == null) {
            methodsMenu = null;
            return;
        }

        // Add the class names to the method definitions. Granted, this
        // is slow, but hopefully there are few classes and few methods
        // in those classes.
        Iterator iter = methods.iterator();
        while (iter.hasNext()) {
            MethodDefinition def = (MethodDefinition) iter.next();
            String cname = findClassForLine(def.getLine());
            def.setClassName(cname);
        }

        // Build a submenu of the sorted method definitions.
        JMenu menu = new JMenu(Bundle.getString(
            "SourceViewPopup.methodsMenuLabel"), true);
        Collections.sort(methods, new MDComparator());
        iter = methods.iterator();
        Preferences prefs = Preferences.userRoot().node(
            "com/bluemarsh/jswat/view");
        boolean shortDesc = prefs.getBoolean(
            "shortMethodDescInPopup", Defaults.VIEW_POPUP_SHORT_METHOD_DESC);
        while (iter.hasNext()) {
            MethodDefinition def = (MethodDefinition) iter.next();
            // Keep the description short, if so desired.
            String desc = shortDesc ? def.getMethodDescShort()
                : def.getMethodDesc();
            if (desc.length() > 80) {
                desc = desc.substring(0, 78) + "...";
            }
            JMenuItem item = new JMenuItem(desc);
            menu.add(item);
            item.setActionCommand(String.valueOf(def.getLine()));
            item.addActionListener(this);
        }
        if (menu.getMenuComponentCount() > 0) {
            methodsMenu = menu;
        }
    } // setClassDefinitions

    /**
     * Set the popup menu items enabled or disabled depending on which
     * line of the source view area the mouse button has been pressed.
     *
     * @param  e        Mouse event.
     * @param  session  Session to operate on.
     */
    protected void setMenuItemsForEvent(MouseEvent e, Session session) {
        // Reset the popup menu by removing all children.
        removeAllItems();

        // Use mouse position to determine line number.
        SourceViewTextArea textArea = (SourceViewTextArea) e.getSource();
        lastClickedLine = textArea.viewToLine(e.getPoint());
        if (lastClickedLine < 0) {
            add((JMenuItem) menuItemTable.get(BAD_LINE_STRING), 0);
            return;
        }

        if (classLines == null) {
            // Parser error when reading source file, thus no class defs.
            add((JMenuItem) menuItemTable.get(PARSER_ERROR_STRING), 0);
            return;
        }

        // Get the class at this line.
        lastClickedClass = findClassForLine(lastClickedLine);
        if (lastClickedClass == null) {
            // No code for this line.
            return;
        }

        // Query existing set of breakpoints for this class and line.
        BreakpointManager bpman = (BreakpointManager)
            session.getManager(BreakpointManager.class);
        lastClickedBreakpoint = bpman.getBreakpoint(
            lastClickedClass, lastClickedLine);

        if (lastClickedBreakpoint == null) {
            // Could not find a breakpoint at this location.
            add((JMenuItem) menuItemTable.get(ADD_STRING), 0);
            add((JMenuItem) menuItemTable.get(RUNTO_STRING), 1);
        } else {
            // Breakpoint exists, add "Breakpoint properties..." item.
            add((JMenuItem) menuItemTable.get(PROPS_STRING), 0);
            // Breakpoint exists, add "Remove breakpoint" item.
            add((JMenuItem) menuItemTable.get(REMOVE_STRING), 1);
            if (lastClickedBreakpoint.isEnabled()) {
                // Breakpoint is enabled, add "Disable breakpoint".
                add((JMenuItem) menuItemTable.get(DISABLE_STRING), 2);
            } else {
                // Breakpoint is disabled, add "Enable breakpoint".
                add((JMenuItem) menuItemTable.get(ENABLE_STRING), 2);
            }
        }
    } // setMenuItemsForEvent

    /**
     * Determine which line the user clicked on and find any breakpoints
     * at that line. If none found, show the "Add breakpoint" menu. If
     * there's a breakpoint, show a popup that provides breakpoint
     * management features.
     *
     * @param  evt  mouse event.
     */
    protected void showPopup(MouseEvent evt) {
        Session session = SessionFrameMapper.getSessionForEvent(evt);
        setMenuItemsForEvent(evt, session);
        show(evt.getComponent(), evt.getX(), evt.getY());
    } // showPopup

    /**
     * Resets the UI property to a value from the current look and feel.
     */
    public void updateUI() {
        super.updateUI();
        // Now update our disconnected children.
        if (methodsMenu != null) {
            methodsMenu.updateUI();
            Component[] children = methodsMenu.getMenuComponents();
            if (children != null && children.length > 0) {
                for (int ii = 0; ii < children.length; ii++) {
                    JComponent jcomp = (JComponent) children[ii];
                    jcomp.updateUI();
                    // We assume this items have no children.
                }
            }
        }
    } // updateUI

    /**
     * Comparator for MethodDefinitions.
     *
     * @author  Neeraj Apte
     */
    private static class MDComparator implements java.util.Comparator {

        /**
         * Comparator implementation used to quickly sort methods based
         * on their signature (excluding the modifiers). The Comparator
         * purposely doesn't throw any exceptions because the
         * functionality is not mission-critical.
         *
         * @param  o1  an <code>Object</code> value.
         * @param  o2  an <code>Object</code> value.
         * @return  an <code>int</code> value.
         */
        public int compare(Object o1, Object o2) {
            if (o1 == null || o2 == null) {
                return (o1 == null && o2 == null) ? 0 : -1;
            } else if (o1 instanceof MethodDefinition
                       && o2 instanceof MethodDefinition) {
                String s1 = ((MethodDefinition) o1).getMethodDescShort();
                String s2 = ((MethodDefinition) o2).getMethodDescShort();
                return s1.compareToIgnoreCase(s2);
            } else {
                return -1;
            }
        } // compare
    } // MDComparator
} // JavaSourceViewPopup
