/*********************************************************************
 *
 *      Copyright (C) 1999-2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      View
 * FILE:        SourceViewPopup.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      04/29/00        Initial version
 *      nf      01/28/01        Changed to use JavaParser results
 *      nf      07/22/01        Allow class defs to be null
 *      nf      08/30/01        Merged action classes into this one
 *      nf      09/01/01        Implemented request 201
 *      nf      09/06/01        Change to support request 236
 *      nf      04/22/02        Implemented RFE 119
 *      nf      04/24/02        Fixed bug 461
 *      nf      04/25/02        Fixed bug 476
 *      nf      07/11/02        Implemented RFE 581
 *      nf      07/19/02        Fixed bug 589
 *      nf      12/26/02        Fixed bug 683
 *      nf      01/30/03        Split out Java-specific code
 *
 * $Id: SourceViewPopup.java 935 2003-04-17 05:00:28Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.view;

import com.bluemarsh.jswat.Defaults;
import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.ui.SessionFrameMapper;
import com.bluemarsh.jswat.ui.SmartPopupMenu;
import com.bluemarsh.jswat.ui.UIAdapter;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.prefs.Preferences;
import javax.swing.JComponent;
import javax.swing.JMenuItem;

/**
 * Class SourceViewPopup defines a popup menu that works specifically for
 * the source view. It constructs a context-sensitive menu based on the
 * existence of breakpoints at the given line of the source view. Like any
 * other popup menu, you must add this popup as a child to the view area in
 * question. It also must be added as a mouse listener to the view area
 * component.
 *
 * @author  Nathan Fiedler
 */
class SourceViewPopup extends SmartPopupMenu implements ActionListener {
    /** Name for launch editor menu item. */
    protected static final String EDITOR_STRING = "launchEditor";
    /** Suffix added to command string to retrieve menu labels. */
    private static final String LABEL_SUFFIX = "Label";
    /** Hashtable of menu items keyed by their names (eg. 'add'). */
    protected Hashtable menuItemTable;
    /** If non-zero, the line the user last right-clicked on. */
    protected int lastClickedLine;
    /** The source file being displayed, if known. */
    protected File sourceFile;

    /**
     * Create a SourceViewPopup.
     *
     * @param  file  file for source, if available.
     */
    public SourceViewPopup(File file) {
        super(Bundle.getString("SourceViewPopup.title"));
        sourceFile = file;
        menuItemTable = new Hashtable();
        menuItemTable.put(EDITOR_STRING, createMenuItem(EDITOR_STRING));
        add((JMenuItem) menuItemTable.get(EDITOR_STRING));
    } // SourceViewPopup

    /**
     * One of the menu items we're listening to was activated.
     *
     * @param  ae  action event.
     */
    public void actionPerformed(ActionEvent ae) {
        // Get the source of the event (it is a JMenuItem).
        JMenuItem menuItem = (JMenuItem) ae.getSource();
        Session session = SessionFrameMapper.getSessionForEvent(ae);
        // Get the action command.
        String cmd = menuItem.getActionCommand();

        if (cmd.equals(EDITOR_STRING)) {
            // Launch the external source editor.
            Preferences prefs = Preferences.userRoot().node(
                "com/bluemarsh/jswat/view");
            String command = prefs.get("extSourceEditor",
                                       Defaults.SOURCE_EDITOR);
            if (command == null || command.length() == 0) {
                session.getUIAdapter().showMessage(
                    UIAdapter.MESSAGE_ERROR,
                    Bundle.getString("SourceViewPopup.noEditorSet"));
            } else if (sourceFile == null) {
                session.getUIAdapter().showMessage(
                    UIAdapter.MESSAGE_ERROR,
                    Bundle.getString("SourceViewPopup.notFile"));
            } else {
                // Replace %f with the filepath and %l with the line number.
                try {
                    int fi = command.indexOf("%f");
                    if (fi >= 0) {
                        command = command.substring(0, fi)
                            + sourceFile.getCanonicalPath()
                            + command.substring(fi + 2);
                    }
                    int li = command.indexOf("%l");
                    if (li >= 0) {
                        command = command.substring(0, li)
                            + lastClickedLine + command.substring(li + 2);
                    }
                    Runtime.getRuntime().exec(command);
                } catch (IOException ioe) {
                    session.getUIAdapter().showMessage(
                        UIAdapter.MESSAGE_ERROR,
                        Bundle.getString("SourceViewPopup.editorError")
                        + ' ' + ioe.getMessage());
                }
            }
        }
    } // actionPerformed

    /**
     * This is the hook through which all menu items are created. Using
     * the <code>cmd</code> string it finds the menu item label in the
     * resource bundle.
     *
     * @param  cmd  name for this menu item, used to get the label
     * @return  new menu item
     */
    protected JMenuItem createMenuItem(String cmd) {
        // Create menu item and set the text label.
        JMenuItem mi = new JMenuItem(Bundle.getString(cmd + LABEL_SUFFIX));
        // Set menu action command.
        mi.setActionCommand(cmd);
        // Set up the action to listen for events.
        mi.addActionListener(this);
        return mi;
    } // createMenuItem

    /**
     * Determine which line the user clicked on and find any breakpoints
     * at that line. If none found, show the "Add breakpoint" menu. If
     * there's a breakpoint, show a popup that provides breakpoint
     * management features.
     *
     * @param  evt  mouse event.
     */
    protected void showPopup(MouseEvent evt) {
        show(evt.getComponent(), evt.getX(), evt.getY());
    } // showPopup

    /**
     * Resets the UI property to a value from the current look and feel.
     */
    public void updateUI() {
        super.updateUI();
        // Now update our disconnected children.
        if (menuItemTable != null) {
            Enumeration enum = menuItemTable.elements();
            while (enum.hasMoreElements()) {
                JComponent jcomp = (JComponent) enum.nextElement();
                jcomp.updateUI();
            }
        }
    } // updateUI
} // SourceViewPopup
