/*********************************************************************
 *
 *      Copyright (C) 2002-2003 Stefano Maestri
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     jEdit JSwat Plugin
 * MODULE:      JSwat Plugin
 * FILE:        BreakpointHighlight.java
 *
 * AUTHOR:      Stefano Maestri
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      sm      03/28/02        Initial version
 *
 * $Id: BreakpointHighlight.java 886 2003-03-10 08:44:29Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.plugins.jedit;

import com.bluemarsh.jswat.breakpoint.Breakpoint;
import com.bluemarsh.jswat.breakpoint.BreakpointEvent;
import com.bluemarsh.jswat.breakpoint.BreakpointListener;
import com.bluemarsh.jswat.breakpoint.BreakpointManager;
import com.bluemarsh.jswat.breakpoint.LocatableBreakpoint;
import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.EBComponent;
import org.gjt.sp.jedit.EBMessage;
import org.gjt.sp.jedit.EditBus;
import org.gjt.sp.jedit.EditPane;
import org.gjt.sp.jedit.msg.BufferUpdate;
import org.gjt.sp.jedit.msg.EditPaneUpdate;
import org.gjt.sp.jedit.textarea.JEditTextArea;
import org.gjt.sp.jedit.textarea.TextAreaExtension;
import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.util.HashMap;
import java.util.Iterator;

/**
 * Class BreakpointHighlight is a text area decorator that indicates the
 * location of breakpoints set by the plugin. This extension acts as a
 * breakpoint listener in order to stay abrest of changes in the
 * breakpoints. It does this in order to maintain a table of lines at
 * which breakpoints are set. This table greatly reduces the time needed
 * in the paintValidLine() method had the table not been utilized.
 *
 * @author  Stefano Maestri
 * @author  Nathan Fiedler
 */
public class BreakpointHighlight extends TextAreaExtension
    implements BreakpointListener, EBComponent {
    /** Color for rendering breakpoints. */
    private static Color breakpointHighlightColor;
    /** Our associated text area. */
    private JEditTextArea textArea;
    /** Is highlight enabled or not. */
    private boolean highlightEnabled;
    /** Table of Integer line numbers and Booleans indicating where
     * breakpoints are located. Lines without breakpoints are not in the
     * table, to conserve memory. Note that the line numbers are
     * zero-based to match jEdit's line numbering scheme. */
    private HashMap linesWithBreakpoints;

    static {
        breakpointHighlightColor = new Color(255, 0, 0);
    }

    /**
     * Constructor for BreakpointHighlight class.
     *
     * @param  textArea  our new text area.
     */
    public BreakpointHighlight(JEditTextArea textArea) {
        this.textArea = textArea;
        linesWithBreakpoints = new HashMap();
    } // BreakpointHighlight

    /**
     * Invoked when a breakpoint has been added.
     *
     * @param  event  breakpoint change event
     */
    public void breakpointAdded(BreakpointEvent event) {
        Buffer buffer = textArea.getBuffer();
        if (buffer.isLoaded()) {
            String classname = JSwatPlugin.getClassForArea(textArea);
            if (classname != null) {
                Breakpoint bp = event.getBreakpoint();
                if (bp instanceof LocatableBreakpoint) {
                    LocatableBreakpoint lbp = (LocatableBreakpoint) bp;
                    if (lbp.matchesClassName(classname)) {
                        linesWithBreakpoints.put(
                            new Integer(lbp.getLineNumber() - 1),
                            Boolean.TRUE);
                    }
                }
            }
        }
    } // breakpointAdded

    /**
     * Invoked when a breakpoint has been changed in some way.
     *
     * @param  event  breakpoint change event
     */
    public void breakpointModified(BreakpointEvent event) {
        // ignored
    } // breakpointModified

    /**
     * Invoked when a breakpoint has been removed.
     *
     * @param  event  breakpoint change event
     */
    public void breakpointRemoved(BreakpointEvent event) {
        Buffer buffer = textArea.getBuffer();
        if (buffer.isLoaded()) {
            String classname = JSwatPlugin.getClassForArea(textArea);
            if (classname != null) {
                Breakpoint bp = event.getBreakpoint();
                if (bp instanceof LocatableBreakpoint) {
                    LocatableBreakpoint lbp = (LocatableBreakpoint) bp;
                    if (lbp.matchesClassName(classname)) {
                        linesWithBreakpoints.remove(
                            new Integer(lbp.getLineNumber() - 1));
                    }
                }
            }
        }
    } // breakpointRemoved

    /**
     * Handle a message from the jEdit edit bus.
     *
     * @param  message  the message.
     */
    public void handleMessage(EBMessage message) {
        if (message instanceof EditPaneUpdate) {
            EditPaneUpdate epu = (EditPaneUpdate) message;
            if (epu.getWhat() == EditPaneUpdate.BUFFER_CHANGED) {
                // Have to re-initialize when the buffer changes.
                init();
            }

        } else if (message instanceof BufferUpdate) {
            BufferUpdate bu = (BufferUpdate) message;
            Buffer buffer = bu.getBuffer();
            if (buffer != textArea.getBuffer()) {
                return;
            }

            if (bu.getWhat() == BufferUpdate.LOADED) {
                // Have to re-initialize when the buffer loads.
                init();

            } else if (bu.getWhat() == BufferUpdate.CLOSED) {
                // The buffer is going away, so are we.
                BreakpointManager breakMgr = (BreakpointManager)
                    JSwatPlugin.getSession().getManager(
                        BreakpointManager.class);
                breakMgr.removeBreakListener(this);
                EditBus.removeFromBus(this);
            }

        }
    } // handleMessage

    /**
     * Initializes the list of breakpoint lines.
     */
    protected void init() {
        linesWithBreakpoints.clear();
        String classname = JSwatPlugin.getClassForArea(textArea);
        if (classname != null) {
            BreakpointManager breakMgr = (BreakpointManager)
                JSwatPlugin.getSession().getManager(
                    BreakpointManager.class);
            Iterator bps = breakMgr.breakpoints(true);

            while (bps.hasNext()) {
                Breakpoint bp = (Breakpoint) bps.next();
                if (bp instanceof LocatableBreakpoint) {
                    LocatableBreakpoint lbp = (LocatableBreakpoint) bp;
                    if (lbp.matchesClassName(classname)) {
                        linesWithBreakpoints.put(
                            new Integer(lbp.getLineNumber() - 1),
                            Boolean.TRUE);
                    }
                }
            }
        }
    } // init

    /**
     * Gets the highlightEnabled attribute.
     *
     * @return  the highlightEnabled value.
     */
    public boolean isHighlightEnabled() {
        return highlightEnabled;
    } // isHighlightEnabled

    /**
     * Paint the decoration on the given line of the text area.
     *
     * @param  gfx           graphics context.
     * @param  screenlLine   line from top of buffer, affected by folding.
     * @param  physicalLine  line from top of file, not affected by folding.
     * @param  start         start of something.
     * @param  end           end of something.
     * @param  y             y coordinates.
     */
    public void paintValidLine(Graphics2D gfx, int screenlLine,
                               int physicalLine, int start, int end, int y) {
        Buffer buffer = textArea.getBuffer();
        if (buffer.isLoaded() && highlightEnabled) {
            Boolean b = (Boolean) linesWithBreakpoints.get(
                new Integer(physicalLine));
            if (b != null) {
                FontMetrics fm = textArea.getPainter().getFontMetrics();
                gfx.setColor(breakpointHighlightColor);
                gfx.fillRect(0, y, textArea.getGutter().getWidth() / 2,
                             fm.getHeight());
            }
        }
    } // paintValidLine

    /**
     * Sets the highlightEnabled attribute.
     *
     * @param  highlightEnabled  the new highlightEnabled value.
     */
    public void setHighlightEnabled(boolean highlightEnabled) {
        this.highlightEnabled = highlightEnabled;
    } // setHighlightEnabled
} // BreakpointHighlight
