/*********************************************************************
 *
 *      Copyright (C) 2001-2003 David Taylor
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     jEdit JSwat Plugin
 * MODULE:      JSwat Plugin
 * FILE:        JEditSourceView.java
 *
 * AUTHOR:      David Taylor
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      dt      2001            Initial version
 *      nf      02/26/03        Minor cleanup
 *      nf      05/08/03        Implement View
 *
 * $Id: JEditSourceView.java 958 2003-05-09 06:59:03Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.plugins.jedit;

import com.bluemarsh.jswat.ContextManager;
import com.bluemarsh.jswat.PathManager;
import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.SourceSource;
import com.bluemarsh.jswat.event.ContextChangeEvent;
import com.bluemarsh.jswat.event.ContextListener;
import com.bluemarsh.jswat.view.View;
import com.sun.jdi.Location;
import com.sun.jdi.ReferenceType;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.io.VFSManager;
import org.gjt.sp.jedit.textarea.JEditTextArea;
import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import javax.swing.JComponent;

/**
 * Defines the JEditSourceView class which will be responsible for
 * displaying the source file on the screen using jEdit buffers.
 *
 * @author  Nathan Fiedler
 * @author  David Taylor
 */
public class JEditSourceView implements ContextListener, View {
    /** The buffer associated with the view. */
    private Buffer theBuffer;
    /** Source file we read from. */
    private File sourceFile;
    /** Session to which we belong. Set in <code>init()</code>. */
    private Session ourSession;
    /** The jEdit view we are associated with. */
    private org.gjt.sp.jedit.View jEditView;

    /**
     * Creates a JEditSourceView object.
     *
     * @param  view    jEdit view to associate with.
     * @param  source  source code file to display.
     */
    public JEditSourceView(org.gjt.sp.jedit.View view, File source) {
        jEditView = view;
        sourceFile = source;

        // Find or create a JEdit buffer with the source file.
        Buffer[] buffers = jEdit.getBuffers();
        for (int ii = 0; ii < buffers.length; ii++) {
            String fname = buffers[ii].getPath();
            File bufferFile = new File(fname);
            if (bufferFile.equals(source)) {
                theBuffer = buffers[ii];
                break;
            }
        }

        if (theBuffer == null) {
            theBuffer = loadFile();
        }
    } // JEditSourceView

    /**
     * Called when the Session is about to begin an active debugging
     * session. That is, JSwat is about to debug a target VM.
     *
     * @param  session  Session being activated.
     */
    public void activate(Session session) {
        // Register for debugger events.
        ContextManager cmgr = (ContextManager)
            session.getManager(ContextManager.class);
        cmgr.addContextListener(this);
    } // activate

    /**
     * This method checks the buffer for this source view is still
     * valid. If not, a new buffer is opened.
     */
    protected void checkBufferValidity() {
        if (theBuffer != null) {
            // ensure buffer is still valid
            Buffer[] bfs = jEdit.getBuffers();
            boolean bufferOk = false;

            for (int ii = 0; ii < bfs.length; ii++) {
                if (theBuffer == bfs[ii]) {
                    bufferOk = true;
                    break;
                }
            }

            if (bufferOk != true) {
                theBuffer = loadFile();
            }
        } else {
            theBuffer = loadFile();
        }

        if (!jEditView.getBuffer().equals(theBuffer)) {
            jEditView.setBuffer(theBuffer);
        }
    } // checkBufferValidity

    /**
     * Called when the Session is about to close down.
     * Closes this view, detaching it from all other objects
     * and preparing it for finalization.
     *
     * @param  session  Session being closed.
     */
    public void close(Session session) {
        ourSession = null;
        sourceFile = null;
    } // close

    /**
     * Invoked when the current context has changed. The context
     * change event identifies which aspect of the context has
     * changed.
     *
     * @param  cce  context change event
     */
    public void contextChanged(ContextChangeEvent cce) {
        ContextManager cmgr = (ContextManager) cce.getSource();
        Location loc = cmgr.getCurrentLocation();
        setCurrentLine(loc);
    } // contextChanged

    /**
     * Called when the Session is about to end an active debugging
     * session. That is, JSwat is about to terminate the connection with
     * the target VM.
     *
     * @param  session  Session being deactivated.
     */
    public void deactivate(Session session) {
        ContextManager cmgr = (ContextManager)
            session.getManager(ContextManager.class);
        cmgr.removeContextListener(this);
        // Erase the current line indicator.
        setCurrentLine(null);
    } // deactivate

    /**
     * Check if the breakpoint is in the same source file as the file
     * we're displaying.
     *
     * @param  location  Location to check.
     * @return  True if breakpoint source name same as our filename.
     */
    protected boolean fileMatch(Location location) {
        // Use source code manager to find location's source file.
        PathManager sourceManager = (PathManager)
            ourSession.getManager(PathManager.class);
        File bpfile = null;

        try {
            ReferenceType type = location.declaringType();
            SourceSource src = sourceManager.mapSource(type);
            bpfile = new File(src.getName());
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }

        if (bpfile != null) {
            return bpfile.equals(sourceFile);
        } else {
            return false;
        }
    } // fileMatch

    /**
     * Look for the given string in this view. Uses the view's current
     * text selection as the starting point. Will wrap around if the
     * string was not found after the current selection.
     *
     * @param  query       string to look for.
     * @param  ignoreCase  true to ignore case.
     * @return  true if string was found somewhere, false if string
     *          does not exist in this view.
     */
    public boolean findString(String query, boolean ignoreCase) {
        // jEdit does this for us.
        return false;
    } // findString

    /**
     * Returns the buffer that this view is associated with.
     *
     * @return  this view's buffer.
     */
    public Buffer getBuffer() {
        return theBuffer;
    } // getBuffer

    /**
     * Returns the long version of title of this view. This may be a
     * file name and path, a fully-qualified class name, or whatever is
     * appropriate for the type of view.
     *
     * @return  long view title.
     */
    public String getLongTitle() {
        // It is not expected for this to be called.
        return "jEdit View";
    } // getLongTitle

    /**
     * Returns the title of this view. This may be a file name, a class
     * name, or whatever is appropriate for the type of view.
     *
     * @return  view title.
     */
    public String getTitle() {
        // It is not expected for this to be called.
        return "jEdit View";
    } // getTitle

    /**
     * Returns a reference to the UI component which can be added to the
     * user interface component tree.
     *
     * @return  interface component.
     */
    public JComponent getUI() {
        // This is not used in jEdit.
        return null;
    } // getUI

    /**
     * Load the source file into a new buffer.
     *
     * @return  newly created buffer.
     */
    protected Buffer loadFile() {
        Buffer buffer = null;
        try {
            buffer = jEdit.openFile(jEditView, sourceFile.getCanonicalPath());
        } catch (IOException ioe) {
            ourSession.getStatusLog().writeStackTrace(ioe);
        }
        return buffer;
    } // loadFile

    /**
     * Called after the Session has added this listener to the
     * Session listener list.
     *
     * @param  session  Session adding this listener.
     */
    public void opened(Session session) {
        ourSession = session;
    } // opened

    /**
     * Read the source data into the text component. The view must be
     * added to the Session as a session listener before calling this
     * method.
     *
     * @param  src   source from which to read data.
     * @param  line  line to make visible.
     * @throws  IOException
     *          if an I/O error occurs in reading the input stream.
     */
    public void refresh(SourceSource src, int line) throws IOException {
        if (theBuffer.isDirty()) {
            theBuffer.reload(jEditView);
        }

        if (ourSession != null) {
            ContextManager cmgr = (ContextManager)
                ourSession.getManager(ContextManager.class);
            setCurrentLine(cmgr.getCurrentLocation());
        }
    } // refresh

    /**
     * Scrolls the source view to the given line, if possible. This will
     * wait for the buffer I/O activity to finish.
     *
     * @param  line  line to scroll to (one-based).
     */
    public void scrollToLine(int line) {
        final int ourLine = (line > 0) ? (line - 1) : 0;

        checkBufferValidity();

        if (theBuffer.isPerformingIO()) {
            VFSManager.runInAWTThread(new Runnable() {
                    public void run() {
                        scrollToLine0(ourLine);
                    }
                });
        } else {
            scrollToLine0(ourLine);
        }
    } // scrollToLine

    /**
     * Does the real work for scrolling to a line.
     *
     * @param  line  line to scroll to.
     */
    private void scrollToLine0(int line) {
        Buffer currentBuffer = jEditView.getBuffer();

        if (currentBuffer == null) {
            jEditView.getEditPane().setBuffer(theBuffer);
        } else if (currentBuffer.equals(theBuffer) != true) {
            jEditView.getEditPane().setBuffer(theBuffer);
        }

        JEditTextArea textArea = jEditView.getTextArea();
        int caret = textArea.getCaretLine();
        textArea.moveCaretPosition(textArea.getLineStartOffset(line));
        textArea.invalidateLine(caret);
    } // scrollToLine0

    /**
     * The the current line indicator using the given location.
     *
     * @param  loc   Current Location
     */
    protected void setCurrentLine(Location loc) {
        // See if this event belongs to our source file.
        if (loc != null && fileMatch(loc)) {
            // Show where the pc is located.
            scrollToLine(loc.lineNumber());
        }
    } // setCurrentLine
} // JEditSourceView
