/*********************************************************************
 *
 *      Copyright (C) 2001-2003 David Taylor
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     jEdit JSwat Plugin
 * MODULE:      JSwat Plugin
 * FILE:        JSwatPlugin.java
 *
 * AUTHOR:      David Taylor
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      dt      2001            Initial version
 *      sm      03/29/02        1.2 version
 *      nf      03/01/03        1.3 version
 *      nf      03/26/03        Fixed bug 737
 *
 * $Id: JSwatPlugin.java 920 2003-04-07 00:18:03Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.plugins.jedit;

import com.bluemarsh.jswat.ContextManager;
import com.bluemarsh.jswat.PathManager;
import com.bluemarsh.jswat.Main;
import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.VMConnection;
import com.bluemarsh.jswat.breakpoint.Breakpoint;
import com.bluemarsh.jswat.breakpoint.BreakpointManager;
import com.bluemarsh.jswat.breakpoint.LineBreakpoint;
import com.bluemarsh.jswat.breakpoint.LocatableBreakpoint;
import com.bluemarsh.jswat.breakpoint.ResolveException;
import com.bluemarsh.jswat.ui.UIAdapter;
import com.bluemarsh.jswat.util.JVMArguments;
import com.bluemarsh.jswat.util.SteppingUtils;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.request.StepRequest;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.EBMessage;
import org.gjt.sp.jedit.EditBus;
import org.gjt.sp.jedit.EditPane;
import org.gjt.sp.jedit.GUIUtilities;
import org.gjt.sp.jedit.MiscUtilities;
import org.gjt.sp.jedit.View;
import org.gjt.sp.jedit.gui.DockableWindowManager;
import org.gjt.sp.jedit.gui.OptionsDialog;
import org.gjt.sp.jedit.msg.BufferUpdate;
import org.gjt.sp.jedit.msg.EditorExiting;
import org.gjt.sp.jedit.msg.EditPaneUpdate;
import org.gjt.sp.jedit.msg.ViewUpdate;
import org.gjt.sp.jedit.textarea.JEditTextArea;
import org.gjt.sp.util.Log;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JOptionPane;

/**
 * The primary interface with the jEdit application.
 *
 * @author  David Taylor
 * @author  Stefano Maestri
 * @author  Nathan Fiedler
 */
public class JSwatPlugin extends org.gjt.sp.jedit.EBPlugin {
    /** Internal name of our plugin. */
    public static final String NAME = "jswatplugin";
    /** Prefix for getting menu properties. */
    public static final String MENU = "jswatplugin.menu";
    /** Prefix for getting options. */
    public static final String OPTION_PREFIX = "options.jswatplugin.";
    /** Prefix for getting properties. */
    private static final String PROPERTY_PREFIX =
        "plugin.jswatplugin.JSwatPlugin.";
    /** The JSwat Session. */
    private static Session theSession;
    /** Map of text areas and fully-qualified class names. */
    private static Map textAreaClassMap;
    /** True if the buffer changed and we need to determine the name of
     * the package in which the class is defined. */
    private static boolean refreshPackageName;

    /**
     * Static initializer for the JSwatPlugin class.
     */
    static {
        textAreaClassMap = new HashMap();
    }

    /**
     * Adds a breakpoint at the current line of the current file.
     *
     * @param  view    associated jEdit view.
     * @param  oneOff  true to expire and delete after hitting once.
     */
    public static void addBreakpoint(View view, boolean oneOff) {
        JEditTextArea textArea = view.getTextArea();
        int line = textArea.getCaretLine() + 1;
        String file = view.getBuffer().getName();

        String classname = getClassForArea(textArea);

        BreakpointManager breakMgr = (BreakpointManager)
            getSession().getManager(BreakpointManager.class);
        try {
            Breakpoint bp = new LineBreakpoint(classname, line);
            breakMgr.addNewBreakpoint(bp);

            if (oneOff) {
                bp.setExpireCount(1);
                bp.deleteOnExpire();
            }
            textArea.invalidateLine(textArea.getCaretLine());
        } catch (ClassNotFoundException cnfe) {
            cnfe.printStackTrace();
        } catch (ResolveException re) {
            re.printStackTrace();
        }
    } // addBreakpoint

    /**
     * Creates the plugin's menu items.
     *
     * @param  menuItems  list of menu items.
     */
    public void createMenuItems(Vector menuItems) {
        menuItems.addElement(GUIUtilities.loadMenu(MENU));
    } // createMenuItems

    /**
     * Create the plugin's options pane.
     *
     * @param  od  options dialog.
     */
    public void createOptionPanes(OptionsDialog od) {
        od.addOptionPane(new JSwatOptionPane());
    } // createOptionPanes

    /**
     * Maps the given text area to a fully-qualified class name. If the
     * name is not yet known, parse the buffer contents to make the
     * determination.
     *
     * @param  textarea  jEdit text area from which to get class name.
     * @return  fully-qualified class name for this buffer.
     */
    public static String getClassForArea(JEditTextArea textarea) {
        Buffer buffer = textarea.getBuffer();
        String classname = (String) textAreaClassMap.get(textarea);
        if (classname == null || refreshPackageName) {
            classname = getFQClassname(textarea);
            if (classname != null) {
                textAreaClassMap.put(textarea, classname);
            }
        }
        return classname;
    } // getClassForArea

    /**
     * Called by jEdit to get the dockable component. We use this to
     * acquire a reference to the associated View.
     *
     * @param  view  View instance.
     * @return  the dockable component.
     */
    public static JComponent getDock(View view) {
        // Add the current stepping line indicator.
        Session session = getSession();
        session.addListener(new SessionToPainterMediator(view));

        // Return our primary interface component.
        JEditUIAdapter adapter = (JEditUIAdapter) getSession().getUIAdapter();
        return adapter.getPanel();
    } // getDock

    /**
     * Returns the fully qualified classname of the current buffer.
     *
     * @param  textArea  the current editing area.
     * @return  the fully qualified classname of the current buffer.
     */
    protected static String getFQClassname(JEditTextArea textArea) {
        String file = textArea.getBuffer().getName();
        String packagename = null;
        String classname = null;

        int lineCount = textArea.getLineCount();
        for (int i = 0; (i < lineCount) && (packagename == null); i++) {
            String lineText = textArea.getLineText(i).trim();
            if (lineText.startsWith("package")) {
                StringTokenizer st = new StringTokenizer(lineText, " \t;");
                int tokenCount = 0;
                while (st.hasMoreTokens() && packagename == null) {
                    String tmp = st.nextToken();
                    tokenCount++;
                    if (tokenCount == 2) {
                        packagename = tmp;
                    }
                }
            }
        }

        if (file != null) {
            int dotIdx = file.indexOf('.');
            if (dotIdx > 0) {
                // -1 or 0 means it is definately not a .java file
                if (packagename != null) {
                    classname = packagename + "."
                        + file.substring(0, file.indexOf('.'));
                } else {
                    classname = file.substring(0, file.indexOf('.'));
                }
                refreshPackageName = false;
            }
        }
        return classname;
    } // getFQClassname

    /**
     * Returns the Session instance. If it does not yet exist, it is
     * created.
     *
     * @return  the session.
     */
    public static Session getSession() {
        if (theSession == null) {
            Main.init();
            Main.setUIAdapter(JEditUIAdapter.class);
            UIAdapter adapter = Main.newUIAdapter();
            if (!Main.detectJPDA(adapter)) {
                // We failed to detect the JPDA.
                throw new IllegalStateException(
                    "Required JPDA (tools.jar) not found.");
            }
            // This assumes only one JSwat session...
            theSession = Main.newSession(adapter);
        }
        return theSession;
    } // getSession

    /**
     * Handle a message from the jEdit core.
     *
     * @param  message  the message.
     */
    public void handleMessage(EBMessage message) {

        if (message instanceof EditPaneUpdate) {
            EditPaneUpdate epu = (EditPaneUpdate) message;
            EditPane editPane = epu.getEditPane();
            JEditTextArea textArea = editPane.getTextArea();

            if (epu.getWhat() == EditPaneUpdate.CREATED) {
                BreakpointHighlight breakpointHighlight =
                    new BreakpointHighlight(textArea);
                textArea.getGutter().addExtension(
                    textArea.getGutter().HIGHEST_LAYER,
                    breakpointHighlight);
                breakpointHighlight.setHighlightEnabled(true);
                EditBus.addToBus(breakpointHighlight);
                BreakpointManager breakMgr = (BreakpointManager)
                    getSession().getManager(BreakpointManager.class);
                breakMgr.addBreakListener(breakpointHighlight);

                VariableValueTooltip vvt = new VariableValueTooltip(textArea);
                textArea.getPainter().addExtension(vvt);

                refreshPackageName = true;
            }

        } else if (message instanceof BufferUpdate) {
            BufferUpdate bu = (BufferUpdate) message;
            if (bu.getWhat() == BufferUpdate.LOADED) {
                refreshPackageName = true;
            }

        } else if (message instanceof ViewUpdate) {
            ViewUpdate vu = (ViewUpdate) message;
            if (vu.getWhat() == ViewUpdate.CREATED) {
                JEditUIAdapter adapter = (JEditUIAdapter)
                    getSession().getUIAdapter();
                adapter.setView(vu.getView());
            }

        } else if (message instanceof EditorExiting) {
            try {
                if (theSession != null) {
                    Main.endSession(theSession);
                    theSession = null;
                }
            } catch (IllegalStateException ise) {
                // do nothing
            }
        }
    } // handleMessage

    /**
     * Display the JSwat help index page.
     *
     * @param  view  view on which to display help.
     */
    public static void helpIndex(View view) {
        JEditUIAdapter adapter = (JEditUIAdapter) getSession().getUIAdapter();
        adapter.showHelp(com.bluemarsh.jswat.Bundle.getResource("helpIndex"));
    } // helpIndex

    /**
     * Remove the breakpoint at the current line.
     *
     * @param  view  associated jEdit view.
     */
    public static void removeBreakpoint(View view) {
        JEditTextArea textArea = view.getTextArea();
        int line = textArea.getCaretLine() + 1;
        String file = view.getBuffer().getName();
        String classname = getClassForArea(textArea);
        BreakpointManager breakMgr = (BreakpointManager)
            getSession().getManager(BreakpointManager.class);
        int physicalLine =
            textArea.getFoldVisibilityManager().virtualToPhysical(line);

        Iterator bps = breakMgr.breakpoints(true);
        while (bps.hasNext()) {
            Breakpoint bp = (Breakpoint) bps.next();
            if (bp instanceof LocatableBreakpoint) {
                LocatableBreakpoint lbp = (LocatableBreakpoint) bp;
                if (lbp.matchesClassName(classname)) {
                    if (lbp.getLineNumber() == physicalLine) {
                        breakMgr.removeBreakpoint(bp);
                        textArea.invalidateLine(textArea.getCaretLine());
                        break;
                    }
                }
            }
        }
    } // removeBreakpoint

    /**
     * Sets the sessionProperties attribute of the JSwatPlugin class.
     */
    private static void setSessionProperties() {
        Session session = getSession();
        session.setProperty("javaHome",
                            jEdit.getProperty(OPTION_PREFIX + "javahome"));
        session.setProperty("jmvExecutable",
                            jEdit.getProperty(OPTION_PREFIX + "javaexec"));
        session.setProperty("sourcepath",
                            jEdit.getProperty(OPTION_PREFIX + "sourcepath"));

        String classPath = jEdit.getProperty(OPTION_PREFIX + "classpath");
        if (classPath != null && classPath.length() > 0) {
            session.setProperty("jvmOptions",
                                jEdit.getProperty(OPTION_PREFIX + "jvmoptions")
                                + " -classpath " + classPath);
        } else {
            session.setProperty("jvmOptions",
                jEdit.getProperty(OPTION_PREFIX + "jvmoptions"));
        }

        String className = jEdit.getProperty(OPTION_PREFIX + "classname");
        if (className != null && className.length() > 0) {
            String appParams = jEdit.getProperty(OPTION_PREFIX + "appparams");
            if (appParams != null && appParams.length() > 0) {
                className = className + " " + appParams;
            }
            session.setProperty("mainClass", className);
        }

        session.setProperty("startSuspended",
                            jEdit.getProperty(OPTION_PREFIX
                                              + "start-suspended"));
    } // setSessionProperties

    /**
     * Show the plugin's options with respect to the given view.
     *
     * @param  view  current view.
     * @return  true to continue, false if user cancelled.
     */
    public static boolean showOptions(View view) {
        JSwatOptionPane options = new JSwatOptionPane();
        options._init();
        Icon icon = new ImageIcon(Bundle.getResource("houseflyImage"));
        int result = JOptionPane.showOptionDialog(
            view, options, "JSwat Options", JOptionPane.OK_CANCEL_OPTION,
            JOptionPane.PLAIN_MESSAGE, icon, null, null);
        if (result == JOptionPane.OK_OPTION) {
            options._save();
            return true;
        } else {
            return false;
        }
    } // showOptions

    /**
     * Perform a single-step operation.
     *
     * @param  size   type of step operation.
     * @param  depth  amount to step.
     */
    protected static void singleStep(int size, int depth) {
        Session session = getSession();
        if (session.isActive()) {
            // Get the current thread.
            ContextManager contextManager = (ContextManager)
                session.getManager(ContextManager.class);
            ThreadReference current = contextManager.getCurrentThread();
            if (current == null) {
                session.getUIAdapter().showMessage(
                    UIAdapter.MESSAGE_ERROR,
                    Bundle.getString("noCurrentThread"));
            } else {
                // Step a single line, into functions.
                if (SteppingUtils.step(
                        session.getVM(), current, size, depth, false,
                        session.getProperty("excludes"))) {
                    // Must use the Session to (quietly) resume the VM.
                    session.resumeVM(JSwatPlugin.class, true, true);
                }
            }
        }
    } // singleStep

    /**
     * Called by jEdit when the plugin is to be started.
     */
    public void start() {
        // If the tools.jar file contains the JPDA classes, this method
        // call will save the user from having to deal with it.
        MiscUtilities.isToolsJarAvailable();
        setSessionProperties();
    } // start

    /**
     * Perform a single-step operation.
     */
    public static void stepInto() {
        singleStep(StepRequest.STEP_LINE, StepRequest.STEP_INTO);
    } // stepInto

    /**
     * Perform a step out operation.
     */
    public static void stepOut() {
        singleStep(StepRequest.STEP_LINE, StepRequest.STEP_OUT);
    } // stepOut

    /**
     * Perform a step over operation.
     */
    public static void stepOver() {
        singleStep(StepRequest.STEP_LINE, StepRequest.STEP_OVER);
    } // stepOver

    /**
     * Called by jEdit when it is about to terminate.
     */
    public void stop() {
        if (theSession != null) {
            Main.endSession(theSession);
            theSession = null;
        }
    } // stop

    /**
     * Start debugging the debuggee. If the session is already active,
     * deactivate it before proceeding.
     *
     * @param  view  the current view.
     */
    public static void vmStart(View view) {
        Session session = getSession();
        if (session.isActive()) {
            session.deactivate(false, JSwatPlugin.class);
        }

        String sd = jEdit.getProperty(OPTION_PREFIX + "show-dialog");
        if (new Boolean(sd).booleanValue()) {
            if (!showOptions(view)) {
                // User cancelled.
                return;
            }
        }

        try {
            DockableWindowManager dwm = view.getDockableWindowManager();
            if (!dwm.isDockableWindowVisible(NAME)) {
                dwm.addDockableWindow(NAME);
            }
            setSessionProperties();
            String className = jEdit.getProperty(OPTION_PREFIX + "classname");
            className = session.getProperty("jvmOptions") + " " + className;

            // Parse the user-provided arguments.
            JVMArguments jvmArgs = new JVMArguments(className);
            String mainClass = jvmArgs.stuffAfterOptions();

            if (mainClass == null || mainClass.length() == 0) {
                // Missing the classname to load.
                throw new IllegalArgumentException("missing class name: "
                                                   + jvmArgs.parsedOptions());
            }

            PathManager pathman = (PathManager)
                session.getManager(PathManager.class);
            String classpath = pathman.getClassPathAsString();
            String jvmOptions = jvmArgs.normalizedOptions(classpath);

            // Build the VM connection with some default values.
            VMConnection connection = VMConnection.buildConnection(
                null, null, jvmOptions, mainClass);

            // Display the options and classname that we're going to use.
            com.bluemarsh.jswat.Log out = session.getStatusLog();
            out.writeln(connection.loadingString());

            // Launch the target VM.
            if (connection.launchDebuggee(session, true)) {
                // Save the start parameters for later reuse by other classes.
                session.setProperty("mainClass", mainClass);
                session.setProperty("jvmOptions", jvmOptions);
                session.setProperty("startSuspended", "true");
            } else {
                out.writeln("vmLoadFailed");
            }
        } catch (Exception e) {
            Log.log(Log.ERROR, JSwatPlugin.class, e);
        }
    } // vmStart

    /**
     * Stop the debuggee and close it.
     */
    public static void vmStop() {
        Session session = getSession();
        if (session.isActive()) {
            session.deactivate(false, JSwatPlugin.class);
        }
    } // vmStop

    /**
     * Resume the debuggee.
     */
    public static void vmResume() {
        Session session = getSession();
        if (session.isActive()) {
            session.resumeVM(JSwatPlugin.class, false, false);
        }
    } // vmResume

    /**
     * Suspend the debuggee.
     */
    public static void vmSuspend() {
        Session session = getSession();
        if (session.isActive()) {
            session.suspendVM(JSwatPlugin.class);
        }
    } // vmSuspend
} // JSwatPlugin
