/*********************************************************************
 *
 *      Copyright (C) 2003 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     jEdit JSwat Plugin
 * MODULE:      JSwat Plugin
 * FILE:        SessionToPainterMediator.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      02/24/03        Initial version
 *
 * $Id: SessionToPainterMediator.java 868 2003-03-01 02:07:08Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.plugins.jedit;

import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.event.SessionEvent;
import com.bluemarsh.jswat.event.SessionListener;
import org.gjt.sp.jedit.View;
import org.gjt.sp.jedit.textarea.TextAreaPainter;

/**
 * Class SessionToPainterMediator ensures that jEdit uses the line
 * highlighter to indicate the current stepping line while JSwat is
 * stepping through code. It does this by ensuring that the line painter
 * is enabled when the session activates, and disabling it when the
 * session deactivates.
 *
 * @author  Nathan Fiedler
 */
public class SessionToPainterMediator implements SessionListener {
    /** The text area painter for the view. */
    private TextAreaPainter originalPainter;
    /** The original line highlighting value. */
    private boolean lineHighlightEnabled;
    /** The View we are associated with. */
    private View ourView;

    /**
     * Constructs a SessionToPainterMediator to work against the given
     * View object.
     *
     * @param  view  the view in which to paint.
     */
    public SessionToPainterMediator(View view) {
        ourView = view;
    } // SessionToPainterMediator

    /**
     * Called when the Session is about to begin an active debugging
     * session. That is, JSwat is about to debug a target VM. Switch on
     * line highlighting so user can see what line the debugger is on.
     *
     * @param  sevt  Session event.
     */
    public void activated(SessionEvent sevt) {
        // This is our current stepping line indicator.
        originalPainter = ourView.getTextArea().getPainter();
        lineHighlightEnabled = originalPainter.isLineHighlightEnabled();
        originalPainter.setLineHighlightEnabled(true);
    } // activated

    /**
     * Called when the Session is about to close down. Restore the line
     * highlighting to user's preference.
     *
     * @param  sevt  Session event.
     */
    public void closing(SessionEvent sevt) {
        if (originalPainter != null) {
            originalPainter.setLineHighlightEnabled(lineHighlightEnabled);
        }
    } // closing

    /**
     * Called when the Session is about to end an active debugging
     * session. That is, JSwat is about to terminate the connection with
     * the target VM. Restore the line highlighting to user's
     * preference.
     *
     * @param  sevt  Session event.
     */
    public void deactivated(SessionEvent sevt) {
        if (originalPainter != null) {
            originalPainter.setLineHighlightEnabled(lineHighlightEnabled);
        }
    } // deactivated

    /**
     * Called after the Session has added this listener to the Session
     * listener list.
     *
     * @param  session  Session adding this listener.
     */
    public void opened(Session session) {
    } // opened

    /**
     * Called when the debuggee is about to be resumed.
     *
     * @param  sevt  session event.
     */
    public void resuming(SessionEvent sevt) {
    } // resuming

    /**
     * Called when the debuggee has been suspended.
     *
     * @param  sevt  session event.
     */
    public void suspended(SessionEvent sevt) {
    } // suspended
} // SessionToPainterMediator
