/*********************************************************************
 *
 *      Copyright (C) 2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Unit Tests
 * FILE:        SessionListenerListTest.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      07/25/02        Initial version
 *
 * $Id: SessionListenerListTest.java 674 2002-11-11 07:16:43Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat;

import com.bluemarsh.jswat.event.SessionEvent;
import com.bluemarsh.jswat.event.SessionListener;
import java.util.ArrayList;
import java.util.List;
import junit.extensions.*;
import junit.framework.*;

/**
 * Tests the SessionListenerList class.
 */
public class SessionListenerListTest extends TestCase {

    public SessionListenerListTest(String name) {
        super(name);
    }

    public static Test suite() {
        return new TestSuite(SessionListenerListTest.class);
    }

    public static void main(String[] args) {
        junit.textui.TestRunner.run(suite());
    }

    public void testSLLBasic() {
        SessionListenerList sll = new SessionListenerList();
        assertNull(sll.acquireListener(-1));
        assertNull(sll.acquireListener(0));
        assertNull(sll.acquireListener(1));
        sll.activateAll(null);
        sll.closeAll(null);
        sll.deactivateAll(null);
        sll.resumeAll(null);
        sll.suspendAll(null);
        assertEquals(0, sll.size());

        try {
            sll.add(null, null, false);
            fail("expected IllegalArgumentException");
        } catch (IllegalArgumentException iae) {
            // this is expected
        }

        try {
            sll.remove(null, null);
            fail("expected IllegalArgumentException");
        } catch (IllegalArgumentException iae) {
            // this is expected
        }

        TestListener l1 = new TestListener();
        sll.add(l1, null, false);
        assertTrue(l1.open);
        assertTrue(!l1.active);
        assertTrue(!l1.running);

        sll.activateAll(null);
        assertTrue(l1.open);
        assertTrue(l1.active);
        assertTrue(!l1.running);

        sll.resumeAll(null);
        assertTrue(l1.open);
        assertTrue(l1.active);
        assertTrue(l1.running);

        sll.suspendAll(null);
        assertTrue(l1.open);
        assertTrue(l1.active);
        assertTrue(!l1.running);

        sll.deactivateAll(null);
        assertTrue(l1.open);
        assertTrue(!l1.active);
        assertTrue(!l1.running);

        sll.closeAll(null);
        assertTrue(!l1.open);
        assertTrue(!l1.active);
        assertTrue(!l1.running);
    }

    public void testSLLThreaded() {
        SessionListenerList sll = new SessionListenerList();
        Simulator sim = new Simulator(sll);

        // populate the list
        List inlist = new ArrayList();
        for (int ii = 0; ii < 100; ii++) {
            TestListener l1 = new TestListener();
            inlist.add(l1);
            sll.add(l1, null, false);
        }

        sim.start();

        List outlist = new ArrayList();
        for (int ii = 0; ii < 100; ii++) {
            TestListener l1 = (TestListener) inlist.remove(
                (int) (Math.random() * inlist.size()));
            outlist.add(l1);
            sll.remove(l1, null);
            Thread.yield();
        }

        for (int ii = 0; ii < 100; ii++) {
            TestListener l1 = (TestListener) outlist.remove(
                (int) (Math.random() * outlist.size()));
            inlist.add(l1);
            sll.add(l1, null, false);
            Thread.yield();
        }

        // the number is arbitrary
        assertTrue(sim.stop() > 100);
    }

    protected class Simulator implements Runnable {
        protected SessionListenerList sll;
        protected Thread runner;
        protected volatile int counter;

        public Simulator(SessionListenerList sll) {
            this.sll = sll;
        }

        public void run() {
            while (!Thread.interrupted()) {
                sll.activateAll(null);
                if (System.currentTimeMillis() % 100 == 0) {
                    // yield the cpu every now and then
                    Thread.yield();
                }
                sll.deactivateAll(null);
                counter++;
            }
        }

        public void start() {
            runner = new Thread(this);
            runner.start();
        }

        public int stop() {
            if (runner != null && runner.isAlive()) {
                runner.interrupt();
                try {
                    runner.join();
                } catch (InterruptedException ie) { }
            }
            return counter;
        }
    }

    protected class TestListener implements SessionListener {
        public boolean open;
        public boolean active;
        public boolean running;

        public void activated(SessionEvent sevt) {
            active = true;
        }

        public void closing(SessionEvent sevt) {
            open = false;
        }

        public void deactivated(SessionEvent sevt) {
            active = false;
        }

        public void opened(Session session) {
            open = true;
        }

        public void resuming(SessionEvent sevt) {
            running = true;
        }

        public void suspended(SessionEvent sevt) {
            running = false;
        }
    }
}
