/*********************************************************************
 *
 *      Copyright (C) 2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Unit Tests
 * FILE:        CommandTestCase.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      07/26/02        Initial version
 *      nf      10/20/02        Avoiding session timing issues
 *
 * $Id: CommandTestCase.java 661 2002-11-02 03:50:31Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.command;

import com.bluemarsh.jswat.CommandManager;
import com.bluemarsh.jswat.Session;
import com.bluemarsh.jswat.event.SessionEvent;
import com.bluemarsh.jswat.event.SessionListener;
import junit.framework.TestCase;

/**
 * Base class for all command test cases.
 *
 * @author  Nathan Fiedler
 */
public class CommandTestCase extends TestCase {

    /**
     * Constructs a command test case with the given name.
     *
     * @param  name  name of test case.
     */
    public CommandTestCase(String name) {
        super(name);
    } // CommandTestCase

    /**
     * Resume the Session and wait for the Session to suspend again.
     * Used when a breakpoint has been set but not yet hit.
     *
     * @param  session  Session to resume.
     * @param  ssl      listener that notices Session stopping.
     */
    protected void resumeAndWait(Session session, SimpleSessionListener ssl) {
        session.resumeVM(this, false, true);
        waitForSuspend(ssl);
    } // resumeAndWait

    /**
     * Get the command and call its perform() method directly.
     *
     * @param  session  open session.
     * @param  input    command input to run.
     */
    protected void runCommand(Session session, String input) {
        CommandManager cm = (CommandManager)
            session.getManager(CommandManager.class);
        CommandArguments cargs = new CommandArguments(input);
        String cname = cargs.nextToken();
        JSwatCommand c = cm.getCommand(cname, false);
        assertNotNull(cname + " is not a command", c);
        c.perform(session, cargs, session.getStatusLog());
    } // runCommand

    /**
     * Runs the given input through the command manager.
     *
     * @param  session  open session.
     * @param  input    command input to run.
     */
    protected void runCommandMgr(Session session, String input) {
        CommandManager cm = (CommandManager)
            session.getManager(CommandManager.class);
        cm.handleInput(input);
    } // runCommandMgr

    /**
     * Wait for the Session to suspend again.
     *
     * @param  ssl  listener that notices Session stopping.
     */
    protected void waitForSuspend(SimpleSessionListener ssl) {
        synchronized (ssl) {
            int count = 0;
            // Give the session a second chance to finish resuming.
            while (ssl.isRunning() && count < 2) {
                try {
                    // wait for the session to suspend
                    ssl.wait();
                } catch (InterruptedException ie) {
                    // ignored
                }
                count++;
            }
            if (!ssl.isActive()) {
                fail("session is not active");
            }
            if (ssl.isRunning()) {
                fail("session is not suspended");
            }
        }
    } // waitForSuspend

    /**
     * Useful for see if the Session has suspended, resumed, etc.
     */
    protected class SimpleSessionListener implements SessionListener {
        /** True if session is open. */
        private volatile boolean open;
        /** True if session is active. */
        private volatile boolean active;
        /** True if session is running. */
        private volatile boolean running;

        /**
         * Session is being activated.
         *
         * @param  sevt  session event.
         */
        public synchronized void activated(SessionEvent sevt) {
            active = true;
            notifyAll();
        }

        /**
         * Session is being closed.
         *
         * @param  sevt  session event.
         */
        public synchronized void closing(SessionEvent sevt) {
            open = false;
            notifyAll();
        }

        /**
         * Session is being deactivated.
         *
         * @param  sevt  session event.
         */
        public synchronized void deactivated(SessionEvent sevt) {
            active = false;
            running = false;
            notifyAll();
        }

        /**
         * Session is being initialized.
         *
         * @param  session  session.
         */
        public synchronized void opened(Session session) {
            open = true;
            notifyAll();
        }

        /**
         * Is session active?
         *
         * @return  true or false.
         */
        public boolean isActive() {
            return active;
        }

        /**
         * Is session open?
         *
         * @return  true or false.
         */
        public boolean isOpen() {
            return open;
        }

        /**
         * Is session running?
         *
         * @return  true or false.
         */
        public boolean isRunning() {
            return running;
        }

        /**
         * Session is resuming.
         *
         * @param  sevt  session event.
         */
        public synchronized void resuming(SessionEvent sevt) {
            running = true;
            notifyAll();
        }

        /**
         * Session is suspending.
         *
         * @param  sevt  session event.
         */
        public synchronized void suspended(SessionEvent sevt) {
            running = false;
            notifyAll();
        }
    }
} // CommandTestCase
