/*********************************************************************
 *
 *      Copyright (C) 2002 Nathan Fiedler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * PROJECT:     JSwat
 * MODULE:      Unit Tests
 * FILE:        StringUtilsTest.java
 *
 * AUTHOR:      Nathan Fiedler
 *
 * REVISION HISTORY:
 *      Name    Date            Description
 *      ----    ----            -----------
 *      nf      07/21/02        Initial version
 *
 * $Id: StringUtilsTest.java 662 2002-11-02 07:25:59Z nfiedler $
 *
 ********************************************************************/

package com.bluemarsh.jswat.util;

import java.util.*;
import junit.extensions.*;
import junit.framework.*;

/**
 * Tests the StringUtils class.
 */
public class StringUtilsTest extends TestCase {

    public StringUtilsTest(String name) {
        super(name);
    }

    public static Test suite() {
        return new TestSuite(StringUtilsTest.class);
    }

    public static void main(String[] args) {
        junit.textui.TestRunner.run(suite());
    }

    public void testStringUtilsCleanForPrinting() {
        // String cleanForPrinting(String input, int maxLen);
        assertEquals("null", StringUtils.cleanForPrinting(null, 0));
        assertEquals("abc123", StringUtils.cleanForPrinting("abc123", 0));
        assertEquals("abc\\n123\\rdef\\t456\\bghi\\f789",
                     StringUtils.cleanForPrinting(
                         "abc\n123\rdef\t456\bghi\f789", 0));
        assertEquals("a long <...>hortened",
                     StringUtils.cleanForPrinting(
                         "a long string that should be shortened", 20));
    }

    public void testStringUtilsIndexOfUnescaped() {
        // int indexOfUnescaped(String s, int ch, int fromIndex);
        assertEquals(-1, StringUtils.indexOfUnescaped(null, 0, 0));
        assertEquals(5, StringUtils.indexOfUnescaped(
                         "a\\'bc'defghi", '\'', 0));
        assertEquals(5, StringUtils.indexOfUnescaped(
                         "a\\'bc'defghi", '\'', 5));
        assertEquals(-1, StringUtils.indexOfUnescaped(
                         "a\\'bc'defghi", '\'', 6));
    }

    public void testStringUtilsListToString() {
        // String listToString(List list);
        assertNull(StringUtils.listToString(null));
        List list = new ArrayList();
        assertEquals("", StringUtils.listToString(list));
        list.add("abc");
        assertEquals("abc", StringUtils.listToString(list));
        list.add("def");
        assertEquals("abc, def", StringUtils.listToString(list));
        list.add("ghi");
        assertEquals("abc, def, ghi", StringUtils.listToString(list));
    }

    public void testStringUtilsSplitOnNewline() {
        // String[] splitOnNewline(String str);
        String[] actual = StringUtils.splitOnNewline(null);
        assertNull(actual);
        actual = StringUtils.splitOnNewline("abcdef");
        assertEquals(1, actual.length);
        assertEquals("abcdef", actual[0]);
        actual = StringUtils.splitOnNewline("abc\ndef");
        assertEquals(2, actual.length);
        assertEquals("abc", actual[0]);
        assertEquals("def", actual[1]);
        actual = StringUtils.splitOnNewline("abc\ndef\nghi");
        assertEquals(3, actual.length);
        assertEquals("abc", actual[0]);
        assertEquals("def", actual[1]);
        assertEquals("ghi", actual[2]);
    }

    public void testStringUtilsStringToList() {
        // List stringToList(String str);
        List list = StringUtils.stringToList(null);
        assertNull(list);
        list = StringUtils.stringToList("");
        assertEquals(0, list.size());
        list = StringUtils.stringToList("abc");
        assertEquals(1, list.size());
        assertEquals("abc", list.get(0));
        list = StringUtils.stringToList("abc,def");
        assertEquals(2, list.size());
        assertEquals("abc", list.get(0));
        assertEquals("def", list.get(1));
        list = StringUtils.stringToList("abc , def , ghi");
        assertEquals(3, list.size());
        assertEquals("abc", list.get(0));
        assertEquals("def", list.get(1));
        assertEquals("ghi", list.get(2));
    }

    public void testStringUtilsToHexString() {
        // String toHexString(int i);
        assertEquals("0001", StringUtils.toHexString(1));
        assertEquals("0080", StringUtils.toHexString(128));
        assertEquals("04d2", StringUtils.toHexString(1234));
        assertEquals("ffff", StringUtils.toHexString(65535));
        assertEquals("0000ffff", StringUtils.toHexString(65535, 8));
        assertEquals("ffffffff", StringUtils.toHexString(-1));
        assertEquals("00000000ffffffff", StringUtils.toHexString(-1, 16));
        assertEquals("000000000000000000000000ffffffff",
                     StringUtils.toHexString(-1, 32));
    }

    public void testStringUtilsTokenize() {
        // String toHexString(int i);
        String[] actual = StringUtils.tokenize(null);
        assertEquals(0, actual.length);
        actual = StringUtils.tokenize("");
        assertEquals(0, actual.length);
        actual = StringUtils.tokenize("abc");
        assertEquals(1, actual.length);
        assertEquals("abc", actual[0]);
        actual = StringUtils.tokenize("abc def");
        assertEquals(2, actual.length);
        assertEquals("abc", actual[0]);
        assertEquals("def", actual[1]);
        actual = StringUtils.tokenize("abc def ghi");
        assertEquals(3, actual.length);
        assertEquals("abc", actual[0]);
        assertEquals("def", actual[1]);
        assertEquals("ghi", actual[2]);
    }

    public void testStringUtilsTrimQuotes() {
        // String trimQuotes(String str);
        assertEquals("abc", StringUtils.trimQuotes("abc"));
        assertEquals("abc", StringUtils.trimQuotes("'abc'"));
        assertEquals("abc'", StringUtils.trimQuotes("abc'"));
        assertEquals("abc", StringUtils.trimQuotes("\"abc\""));
        assertEquals("abc\"", StringUtils.trimQuotes("abc\""));
        assertEquals("'abc", StringUtils.trimQuotes("'abc"));
        assertEquals("\"abc", StringUtils.trimQuotes("\"abc"));
    }
}
